import os

import pandas as pd
from pandas import DataFrame

from constants.structure import DOCKER_FOLDER_DATA, VCF_MOUNTED_DOCKER
from database.Execution import Execution
from enums.MetadataColumns import MetadataColumns
from enums.Profile import Profile
from enums.VcfColumns import vcf_columns
from utils.setup_logger import log


class Preprocess:
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str):
        self.execution = execution
        self.data = data
        self.metadata = metadata
        self.profile = profile

    def run(self):
        self.preprocess()
        self.add_vcf_files_in_data()

    def preprocess(self):
        pass

    def add_vcf_files_in_data(self) -> None:
        if self.profile == Profile.GENOMIC:
            mapping_pid_vcf = []
            pid_column_name = os.getenv("PATIENT_ID")  # the data is not normalized yet, so we keep the non-normalized column name (not self.execution.patient_id_column_name)
            filepath_column_name = vcf_columns[self.execution.hospital_name]
            if filepath_column_name is not None:
                for entry in os.getenv("DATA_FILES").split(","):
                    if "*.vcf" in entry:
                        # this is the directory which contains all the VCF files
                        the_entry = entry.replace("*.vcf", "")  # os.listdir requires the folder name (VCF-FILES/) without the specification
                        log.info(the_entry)
                        if the_entry == "":
                            the_entry = "."  # the VCF files are next to the data files
                        for vcf_file in os.listdir(os.path.join(DOCKER_FOLDER_DATA, the_entry)):
                            if ".vcf" in vcf_file:
                                pid = vcf_file.replace(".vcf", "")
                                mapping_pid_vcf.append({pid_column_name: pid, filepath_column_name: os.path.join(VCF_MOUNTED_DOCKER, vcf_file)})
                            else:
                                log.info(f"skip non VCF file {vcf_file}")
                    else:
                        # log.info(f"Skip {entry}")
                        pass
                pid_vcf_df = DataFrame(mapping_pid_vcf)
                self.data = self.data.merge(pid_vcf_df, on=pid_column_name, how="outer")
            else:
                log.info("No name provided for the VCF file column. Skipping it.")

    @classmethod
    def get_subset_of_columns_in_df(cls, df: DataFrame, file_type: Profile, metadata: DataFrame) -> DataFrame:
        profile_filename = Profile.get_preprocess_data_filename(filetype=file_type)
        columns = metadata[metadata[MetadataColumns.PROFILE] == profile_filename][MetadataColumns.COLUMN_NAME]
        columns = [MetadataColumns.normalize_name(column_name) for column_name in columns]
        # df_samples = df_samples[sample_columns]  # nope, this raises an error if some of the columns to keep do not exist in the data
        return df.loc[:, pd.Index(columns).intersection(df.columns)]



import os

import pandas as pd
from pandas import DataFrame

from constants.structure import DOCKER_FOLDER_DATA
from database.Execution import Execution
from enums.Profile import Profile
from preprocessing.Preprocess import Preprocess
from utils.file_utils import read_tabular_file_as_string


def get_sample_number(x):
    extracted_number = x[str(x).rfind("_") + 1:len(str(x))]
    try:
        return int(extracted_number)
    except:
        return 0


class PreprocessKidneyCovid(Preprocess):
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str):
        super().__init__(execution=execution, data=data, metadata=metadata, profile=profile)

    def preprocess(self):
        if self.profile == Profile.CLINICAL or self.profile == Profile.GENOMIC:
            df_barcode_to_patient = read_tabular_file_as_string(os.path.join(DOCKER_FOLDER_DATA, "mapping_patient_sample.csv"))
            df_barcode_to_patient = df_barcode_to_patient[["sample_id", "individual_id"]]
            df_barcode_to_patient = df_barcode_to_patient.drop_duplicates()

        if self.profile in [Profile.PHENOTYPIC, Profile.IMAGING, Profile.DIAGNOSIS]:
            # for metadata_w1.csv
            if self.profile == Profile.PHENOTYPIC:
                # process samples data to transpose them
                self.data = self.data.drop(["cause_eskd", "WHO_severity", "WHO_temp_severity", "fatal_disease", "case_control", "radiology_evidence_covid", "time_from_first_symptoms", "time_from_first_positive_swab"], axis=1)
            elif self.profile == Profile.IMAGING:
                self.data = self.data[["individual_id", "sample_id", "radiology_evidence_covid"]]
            elif self.profile == Profile.DIAGNOSIS:
                self.data = self.data[["individual_id", "sample_id", "cause_eskd", "WHO_severity", "fatal_disease"]]
            # after selecting columns of interest, we compute the latest sample id of the patient and keep associated data
            sample_numbers = self.data["sample_id"].apply(get_sample_number)
            self.data = self.data.assign(sample_number=sample_numbers.values)  # assign values to column 'sample_number'
            self.data["sample_max"] = self.data.groupby(["individual_id"])["sample_number"].transform("max")
            self.data = self.data[self.data["sample_number"] == self.data["sample_max"]]
            self.data = self.data.drop(["sample_id", "sample_number", "sample_max"], axis="columns")
            self.data = self.data.reset_index(drop=True)  # do not the column "index", simply re-number rows

        elif self.profile == Profile.CLINICAL:
            # for general_panel.csv
            # we need to associate each sample barcode to the corresponding patient ID
            self.data = self.data.rename({"Sample_ID": "sample_id"}, axis="columns")
            self.data = pd.merge(self.data, df_barcode_to_patient, on="sample_id", how="left")

        elif self.profile == Profile.GENOMIC:
            # the htseq_counts.csv file has been processed beforehand to select a smaller subset of genes and
            # to attach the individual id to each sequence count
            pass




from enums.EnumAsClass import EnumAsClass
from utils.setup_logger import log


class TableNames(EnumAsClass):
    HOSPITAL = "Hospital"
    PATIENT = "Patient"
    FEATURE = "Feature"
    RECORD = "Record"
    TOP10_FEATURES = "TMP_Top10Features"
    PAIRS_FEATURES = "TMP_PairsFeatures"
    COUNTS_PATIENTS = "TMP_CountsPatients"
    COUNTS_SAMPLES = "TMP_CountsSamples"
    COUNTS_FEATURES = "TMP_CountsFeatures"
    FEATURE_PROFILE = "FeatureProfile"
    # PHENOTYPIC_FEATURE = "PhenotypicFeature"
    # PHENOTYPIC_RECORD = "PhenotypicRecord"
    # CLINICAL_FEATURE = "ClinicalFeature"
    # CLINICAL_RECORD = "ClinicalRecord"
    # DIAGNOSIS_FEATURE = "DiagnosisFeature"
    # DIAGNOSIS_RECORD = "DiagnosisRecord"
    # GENOMIC_FEATURE = "GenomicFeature"
    # GENOMIC_RECORD = "GenomicRecord"
    # MEDICINE_FEATURE = "MedicineFeature"
    # MEDICINE_RECORD = "MedicineRecord"
    # IMAGING_FEATURE = "ImagingFeature"
    # IMAGING_RECORD = "ImagingRecord"
    EXECUTION = "Execution"
    STATS_DB = "DatabaseStatistics"
    STATS_TIME = "TimeStatistics"
    STATS_QUALITY = "QualityStatistics"
    TEST = "Test"
    DATASET = "Dataset"
    VIEW_FEATURES_DATASET = "ViewFeaturesDatasets"

    # IMPORTANT NOTE:
    # do NOT import Database for type hinting in methods defined here
    # otherwise, this creates a circular dependency between Database and TableNames

    @classmethod
    def data_tables(cls) -> list:
        return [TableNames.HOSPITAL, TableNames.PATIENT, TableNames.RECORD, TableNames.FEATURE, TableNames.DATASET]

    @classmethod
    def values(cls, db):
        # override the values() method of Enum to only return tables that exists in the DB
        table_names = []
        for name, value in vars(cls).items():
            if not (name.startswith('__') or isinstance(value, classmethod)) and (db is None or (db.check_table_exists(table_name=value) and db is not None)):
                # if DB is None, we do not want to check particularly whether the tables exist or not (we simply want to iterate over the table names)
                # if DB is not None, we want to check that the table exists in the DB
                if "Statistics" not in value and "Execution" not in value:
                    # this is not a table for reporting stats
                    table_names.append(value)
        log.info(table_names)
        return table_names



import dataclasses

from entities.Feature import Feature
from enums.Profile import Profile
from enums.TableNames import TableNames


@dataclasses.dataclass()
class ImagingFeature(Feature):
    entity_type: str = f"{Profile.IMAGING}{TableNames.FEATURE}"



from enums.EnumAsClass import EnumAsClass


class AccessTypes(EnumAsClass):
    USER_AGENT = "user_agent"
    AUTHENTICATION = "authentication"
    API_KEY_IN_HEADER = "api_key_in_header"
    API_KEY_IN_BEARER = "api_key_in_bearer"
    API_KEY_IN_URL = "api_key_in_url"



from enums.EnumAsClass import EnumAsClass
from utils.setup_logger import log
from utils.str_utils import process_spaces


class DataTypes(EnumAsClass):
    INTEGER = "int"
    FLOAT = "float"
    STRING = "str"
    DATE = "date"
    DATETIME = "datetime"
    BOOLEAN = "bool"
    CATEGORY = "category"
    IMAGE = "image"
    REGEX = "regex"
    API = "api"
    LIST = "list"

    METADATA_NB_UNRECOGNIZED_ETL_TYPE = 0

    @classmethod
    def normalize(cls, data_type: str) -> str:
        if data_type == "":
            return ""
        else:
            data_type = process_spaces(input_string=data_type)
            data_type = data_type.lower()

            if data_type in ["int", "integer"]:
                return DataTypes.INTEGER
            elif data_type in ["str", "string"]:
                return DataTypes.STRING
            elif data_type in ["category", "categorical"]:
                return DataTypes.CATEGORY
            elif data_type in ["float", "numeric"]:
                return DataTypes.FLOAT
            elif data_type in ["bool", "boolean"]:
                return DataTypes.BOOLEAN
            elif data_type == "image file":
                return DataTypes.IMAGE
            elif data_type == "date":
                return DataTypes.DATE
            elif data_type in ["datetime", "datetime64"]:
                return DataTypes.DATETIME
            elif data_type == "regex":
                return DataTypes.REGEX
            elif data_type == "api":
                return DataTypes.API
            elif data_type == "list":
                return DataTypes.LIST
            else:
                log.error(f"{data_type} is not a recognized data type; we will use string type by default.")
                DataTypes.METADATA_NB_UNRECOGNIZED_ETL_TYPE += 1
                return DataTypes.STRING

    @classmethod
    def numeric(cls) -> list:
        return [DataTypes.INTEGER, DataTypes.FLOAT]

    @classmethod
    def categorical(cls) -> list:
        return [DataTypes.STRING, DataTypes.BOOLEAN, DataTypes.CATEGORY, DataTypes.API, DataTypes.LIST]

    @classmethod
    def dates(cls) -> list:
        return [DataTypes.DATE, DataTypes.DATETIME]



import os

WORKING_DIR = "working-dir"
DEFAULT_DB_NAME = "better_default"
# these constants have to exactly match the volume paths described in compose.yaml
DOCKER_FOLDER = "/home/i-etl-deployed"
DOCKER_FOLDER_DATA = os.path.join(DOCKER_FOLDER, "real-data")
DOCKER_FOLDER_METADATA = os.path.join(DOCKER_FOLDER, "metadata")
# DOCKER_FOLDER_ANONYMIZED_PATIENT_IDS = os.path.join(DOCKER_FOLDER, "pids")
DOCKER_FOLDER_TEST = os.path.join(DOCKER_FOLDER, "datasets", "test")
VCF_MOUNTED_DOCKER = "/home/data"
DB_CONNECTION = "mongodb://mongo:27017/"

TEST_DB_NAME = "better_test"



from __future__ import annotations

import re
from datetime import datetime
from itertools import islice
from typing import Any

import pandas as pd
import ujson
from pandas import DataFrame

from constants.defaults import BATCH_SIZE, PATTERN_VALUE_UNIT, NO_ID
from database.Counter import Counter
from database.Database import Database
from entities.Dataset import Dataset
from database.Execution import Execution
from entities.ClinicalFeature import ClinicalFeature
from entities.ClinicalRecord import ClinicalRecord
from entities.DiagnosisFeature import DiagnosisFeature
from entities.DiagnosisRecord import DiagnosisRecord
from entities.Feature import Feature
from entities.GenomicFeature import GenomicFeature
from entities.GenomicRecord import GenomicRecord
from entities.Hospital import Hospital
from entities.ImagingFeature import ImagingFeature
from entities.ImagingRecord import ImagingRecord
from entities.MedicineFeature import MedicineFeature
from entities.MedicineRecord import MedicineRecord
from entities.OntologyResource import OntologyResource
from entities.Patient import Patient
from entities.PhenotypicFeature import PhenotypicFeature
from entities.PhenotypicRecord import PhenotypicRecord
from entities.Resource import Resource
from enums.DataTypes import DataTypes
from enums.DiagnosisColumns import DiagnosisColumns
from enums.Domain import Domain
from enums.MetadataColumns import MetadataColumns
from enums.Ontologies import Ontologies
from enums.Profile import Profile
from enums.TableNames import TableNames
from enums.TimerKeys import TimerKeys
from enums.Visibility import Visibility
from etl.Task import Task
from src.constants.defaults import DEFAULT_NAN_VALUE
from main_statistics.QualityStatistics import QualityStatistics
from main_statistics.TimeStatistics import TimeStatistics
from utils.cast_utils import cast_str_to_boolean, cast_str_to_datetime, cast_str_to_float, cast_str_to_int
from utils.file_utils import write_in_file
from utils.setup_logger import log


class Transform(Task):

    def __init__(self, database: Database, execution: Execution, data: DataFrame, metadata: DataFrame,
                 mapping_column_to_categorical_value: dict,
                 mapping_column_to_unit: dict, mapping_column_to_domain: dict,
                 mapping_column_to_type: dict | None,
                 profile: str, dataset_id: int, dataset_key: Dataset, load_patients: bool,
                 quality_stats: QualityStatistics):
        super().__init__(database=database, execution=execution, quality_stats=quality_stats)
        self.time_statistics = TimeStatistics(record_stats=True)
        self.counter = Counter()  # resource counter
        self.profile = profile
        self.load_patients = load_patients
        self.dataset_id = dataset_id  # file number (one for each dataset)
        self.dataset_instance = dataset_key

        # get data, metadata and the mapped values computed in the Extract step
        self.data = data
        self.metadata = metadata
        self.mapping_column_to_unit = mapping_column_to_unit
        self.mapping_column_to_categorical_value = mapping_column_to_categorical_value
        if mapping_column_to_type is not None:
            self.mapping_column_to_type = mapping_column_to_type  # we give it to Transform during tests
        else:
            self.mapping_column_to_type = {}
        self.mapping_column_to_visibility = {}
        self.mapping_column_to_domain = mapping_column_to_domain
        self.mapping_apivalue_to_onto_resource = {}  # for API columns only; of the form: <"onto_name:onto_code": onto resource>
        # to keep track of anonymized vs. hospital patient ids
        # this is empty if no file as been provided by the user, otherwise it contains some mappings <patient ID, anonymized ID>
        self.patient_ids_mapping = {}
        # to keep track of the total number of values that could exist (whether they are Nan or a real value)
        self.mapping_column_all_count = {}

        # to record objects that will be further inserted in the database
        self.features = []
        self.records = []
        self.hospitals = []
        self.patients = []
        self.samples = []

    def run(self) -> None:
        self.load_patient_id_mapping()  # we always load the mapping in order to retrieve existing identifiers when creating data for existing patients
        if self.load_patients:
            # this is the first profile of the dataset, we load patients
            log.info("********** create patients")
            self.counter.set_with_database(database=self.database)
            self.create_patients()
        else:
            # this is another profile of the same dataset, we do not reload the same patients
            pass

        log.info(f"********** create {self.profile} features and records")
        self.counter.set_with_database(database=self.database)
        self.create_features()
        self.counter.set_with_database(database=self.database)
        self.create_records()

    ##############################################################
    # FEATURES
    ##############################################################

    def create_features(self) -> None:
        # 1. get existing features in memory
        log.info(f"Retrieving {self.profile}Feature from database")
        result = self.database.find_operation(table_name=TableNames.FEATURE, filter_dict={Resource.ENTITY_TYPE_: f"{self.profile}Feature"}, projection={Feature.NAME_: 1, Resource.IDENTIFIER_: 1, Resource.DATASET_: 1})
        db_existing_features = [res[Feature.NAME_] for res in result]  # list of features (names) existing in this dataset
        log.info(db_existing_features)

        # 2. create non-existing features in-memory, then insert them
        log.info(f"Creating {self.profile}Feature instances in memory")
        columns = self.metadata.columns
        for row in self.metadata.itertuples(index=False):
            column_name = row[columns.get_loc(MetadataColumns.COLUMN_NAME)]
            # columns to remove have already been removed in the Extract part from the metadata
            # here, we need to ensure that we create Features that have a name, which are not IDs (patient or sample) nor for Diagnosis counter (clinical base_id)
            if column_name not in ["", self.execution.patient_id_column_name, self.execution.sample_id_column_name, DiagnosisColumns.DISEASE_COUNTER]:
                if column_name not in db_existing_features:
                    # we create a new Feature from scratch
                    onto_resource = self.create_ontology_resource_from_row(column_name=column_name)
                    data_type = row[columns.get_loc(MetadataColumns.ETL_TYPE)]  # this has been normalized while loading + we take ETL_type to get the narrowest type (in which we cast values)
                    if data_type == DataTypes.API:  # because the catalogue does not know about the API datatype; each API call leads to an ontology resource, thus a category
                        data_type = DataTypes.CATEGORY
                    self.mapping_column_to_type[column_name] = data_type
                    visibility = row[columns.get_loc(MetadataColumns.VISIBILITY)]  # this has been normalized while loading
                    self.mapping_column_to_visibility[column_name] = visibility
                    unit = self.mapping_column_to_unit[column_name] if column_name in self.mapping_column_to_unit else None  # else covers: there is no dataType for this column; there is no datatype in that type of entity
                    description = row[columns.get_loc(MetadataColumns.SIGNIFICATION_EN)]
                    normalized_categorical_values = None
                    domain = {}
                    if data_type in [DataTypes.CATEGORY, DataTypes.REGEX] and column_name in self.mapping_column_to_categorical_value:
                        # for categorical values, we first need to take the list of (normalized) values that are available for the current column, and then take their CC
                        # this avoids to add categorical values for boolean features (where Yes and No and encoded with ontology resource), we do not add them
                        normalized_categorical_values = list(self.mapping_column_to_categorical_value[column_name].values())
                        log.info(normalized_categorical_values)
                        # categorical_values = [self.mapping_categorical_value_to_onto_resource[normalized_categorical_value] for normalized_categorical_value in normalized_categorical_values]
                        domain[Domain.ACCEPTED_VALUES] = list(self.mapping_column_to_categorical_value[column_name].keys())
                        log.info(domain[Domain.ACCEPTED_VALUES])
                    elif data_type in [DataTypes.DATE, DataTypes.DATETIME] or data_type in DataTypes.numeric():
                        if column_name in self.mapping_column_to_domain and self.mapping_column_to_domain[column_name] is not None:
                            if Domain.MIN in self.mapping_column_to_domain[column_name]:
                                domain[Domain.MIN] = self.mapping_column_to_domain[column_name][Domain.MIN]
                            if Domain.MAX in self.mapping_column_to_domain[column_name]:
                                domain[Domain.MAX] = self.mapping_column_to_domain[column_name][Domain.MAX]
                    if self.profile == Profile.PHENOTYPIC:
                        new_feature = PhenotypicFeature(identifier=NO_ID, name=column_name,
                                                        ontology_resource=onto_resource,
                                                        data_type=data_type, unit=unit,
                                                        counter=self.counter,
                                                        categories=normalized_categorical_values,
                                                        visibility=visibility,
                                                        dataset=self.dataset_instance.global_identifier,
                                                        description=description,
                                                        domain=domain)
                    elif self.profile == Profile.CLINICAL:
                        new_feature = ClinicalFeature(identifier=NO_ID, name=column_name, ontology_resource=onto_resource,
                                                      data_type=data_type, unit=unit,
                                                      counter=self.counter,
                                                      categories=normalized_categorical_values,
                                                      visibility=visibility,
                                                      dataset=self.dataset_instance.global_identifier,
                                                      description=description,
                                                      domain=domain)
                    elif self.profile == Profile.DIAGNOSIS:
                        new_feature = DiagnosisFeature(identifier=NO_ID, name=column_name,
                                                       ontology_resource=onto_resource,
                                                       data_type=data_type,
                                                       unit=unit, counter=self.counter,
                                                       categories=normalized_categorical_values,
                                                       visibility=visibility,
                                                       dataset=self.dataset_instance.global_identifier,
                                                       description=description,
                                                       domain=domain)
                    elif self.profile == Profile.GENOMIC:
                        new_feature = GenomicFeature(identifier=NO_ID, name=column_name,
                                                     ontology_resource=onto_resource,
                                                     data_type=data_type,
                                                     unit=unit, counter=self.counter,
                                                     categories=normalized_categorical_values,
                                                     visibility=visibility,
                                                     dataset=self.dataset_instance.global_identifier,
                                                     description=description,
                                                     domain=domain)
                    elif self.profile == Profile.IMAGING:
                        new_feature = ImagingFeature(identifier=NO_ID, name=column_name,
                                                     ontology_resource=onto_resource,
                                                     data_type=data_type,
                                                     unit=unit, counter=self.counter,
                                                     categories=normalized_categorical_values,
                                                     visibility=visibility,
                                                     dataset=self.dataset_instance.global_identifier,
                                                     description=description,
                                                     domain=domain)
                    elif self.profile == Profile.MEDICINE:
                        new_feature = MedicineFeature(identifier=NO_ID, name=column_name,
                                                      ontology_resource=onto_resource,
                                                      data_type=data_type,
                                                      unit=unit, counter=self.counter,
                                                      categories=normalized_categorical_values,
                                                      visibility=visibility,
                                                      dataset=self.dataset_instance.global_identifier,
                                                      description=description,
                                                      domain=domain)
                    else:
                        raise NotImplementedError("To be implemented")

                    if onto_resource is not None:
                        log.info(f"adding a new {self.profile} feature about {onto_resource.label}: {new_feature}")
                    else:
                        # no associated ontology code or failed to retrieve the code with API
                        log.info(f"adding a new {self.profile} feature about {column_name}: {new_feature}")

                    self.features.append(new_feature.to_json())  # this cannot be null, otherwise we would have raise the above exception
                    if len(self.features) >= BATCH_SIZE:
                        self.process_batch_of_features()
                else:
                    # the Feature already exists, so no need to add it to the database again.
                    # however, we need to update the set of datasets in which it appears
                    log.error(f"The feature about {column_name} already exists. Not added.")
            else:
                log.debug(f"I am skipping column {column_name} because it has been dropped or is an ID column or its name is empty.")
        # save the remaining tuples that have not been saved (because there were less than BATCH_SIZE tuples before the loop ends).
        if len(self.features) > 0:
            self.process_batch_of_features()
        # write all the features in the database now (to be able to retrieve them for creating records just after
        self.database.load_json_in_table(table_name=TableNames.FEATURE, unique_variables=[Feature.NAME_], dataset_id=self.dataset_id)

    ##############################################################
    # RECORDS
    ##############################################################

    def create_records(self) -> None:
        log.info(f"creating {self.profile}Record instances in memory")

        # a. load some data from the database to compute references
        mapping_hospital_to_hospital_id = self.database.retrieve_mapping(table_name=TableNames.HOSPITAL,
                                                                         key_fields=Hospital.NAME_,
                                                                         value_fields=Resource.IDENTIFIER_,
                                                                         filter_dict={})
        log.info(mapping_hospital_to_hospital_id)
        mapping_column_to_feature_id = self.database.retrieve_mapping(table_name=TableNames.FEATURE,
                                                                      key_fields=Feature.NAME_,
                                                                      value_fields=Resource.IDENTIFIER_,
                                                                      filter_dict={Resource.ENTITY_TYPE_: f"{self.profile}{TableNames.FEATURE}"})
        log.info(f"{len(mapping_column_to_feature_id)} {self.profile}{TableNames.FEATURE} have been retrieved from the database.")
        # if len(mapping_column_to_feature_id) > 10:
        #     # print only the ten first elements
        #     log.info(dict(islice(mapping_column_to_feature_id.items(), 10)))
        #
        # else:
        # log.info(mapping_column_to_feature_id)

        # b. Create Record instance, and write them in temporary (JSON) files
        columns = self.data.columns
        for row in self.data.itertuples(index=False):
            # log.info(row)
            # create Record instances by associating observations to a patient, a record and a hospital
            for column_name in columns:
                value = row[columns.get_loc(column_name)]
                # log.debug(f"column {column_name} (type: {type(column_name)}), value is {value}")
                if value == "":
                    # if there is no value for that Feature, no need to create a Record instance
                    # log.error(f"skipping value {value} in column {column_name} because it is None, or empty or nan")
                    self.quality_stats.count_empty_cell_for_column(column_name=column_name)
                    if column_name in mapping_column_to_feature_id:
                        feature_id = mapping_column_to_feature_id[column_name]
                        if feature_id not in self.mapping_column_all_count:
                            self.mapping_column_all_count[feature_id] = 1
                        else:
                            self.mapping_column_all_count[feature_id] = self.mapping_column_all_count[feature_id] + 1
                else:
                    if column_name in mapping_column_to_feature_id:
                        # we know a code for this column, so we can register the value of that Feature in a new Record
                        feature_id = mapping_column_to_feature_id[column_name]
                        hospital_id = mapping_hospital_to_hospital_id[self.execution.hospital_name]
                        # get the anonymized patient id using the mapping <initial id, anonymized id>
                        patient_id = self.patient_ids_mapping[row[columns.get_loc(self.execution.patient_id_column_name)]]
                        fairified_value = self.fairify_value(column_name=column_name, value=value)
                        anonymized_value, is_anonymized = self.anonymize_value(column_name=column_name,
                                                                               fairified_value=fairified_value)
                        if is_anonymized:
                            fairified_value = anonymized_value  # we could anonymize this value, this is the one to insert in the DB
                        dataset = self.execution.current_dataset_gid
                        if self.profile == Profile.PHENOTYPIC:
                            new_record = PhenotypicRecord(identifier=NO_ID,
                                                          instantiates=feature_id,
                                                          has_subject=patient_id,
                                                          registered_by=hospital_id,
                                                          value=fairified_value,
                                                          counter=self.counter,
                                                          dataset=dataset)
                        elif self.profile == Profile.CLINICAL:
                            if self.execution.sample_id_column_name in columns:
                                # this dataset contains a sample barcode (or equivalent)
                                base_id = row[columns.get_loc(self.execution.sample_id_column_name)]
                            else:
                                base_id = None
                            new_record = ClinicalRecord(identifier=NO_ID,
                                                        instantiates=feature_id,
                                                        has_subject=patient_id,
                                                        registered_by=hospital_id,
                                                        value=fairified_value,
                                                        base_id=base_id,
                                                        counter=self.counter,
                                                        dataset=dataset)
                        elif self.profile == Profile.DIAGNOSIS:
                            if DiagnosisColumns.DISEASE_COUNTER in columns:
                                # this dataset contains a diagnosis counter because patients may be affected
                                # by several diseases
                                # we also need to force the conversion to int, because we read the data as str
                                # and such values do not go through the fairification method
                                try:
                                    diagnosis_counter = int(row[columns.get_loc(DiagnosisColumns.DISEASE_COUNTER)])
                                except:
                                    # the value is None because the patient diseases is unknown in the disease classification
                                    diagnosis_counter = None
                            else:
                                diagnosis_counter = None
                            new_record = DiagnosisRecord(identifier=NO_ID,
                                                         instantiates=feature_id,
                                                         has_subject=patient_id,
                                                         registered_by=hospital_id,
                                                         value=fairified_value,
                                                         diagnosis_counter=diagnosis_counter,
                                                         counter=self.counter,
                                                         dataset=dataset)
                        elif self.profile == Profile.GENOMIC:
                            new_record = GenomicRecord(identifier=NO_ID,
                                                       instantiates=feature_id,
                                                       has_subject=patient_id,
                                                       registered_by=hospital_id,
                                                       vcf=None,
                                                       value=fairified_value,
                                                       counter=self.counter,
                                                       dataset=dataset)
                        elif self.profile == Profile.IMAGING:
                            new_record = ImagingRecord(identifier=NO_ID,
                                                       instantiates=feature_id,
                                                       has_subject=patient_id,
                                                       registered_by=hospital_id,
                                                       scan=None,
                                                       value=fairified_value,
                                                       counter=self.counter,
                                                       dataset=dataset)
                        elif self.profile == Profile.MEDICINE:
                            new_record = MedicineRecord(identifier=NO_ID,
                                                        instantiates=feature_id,
                                                        has_subject=patient_id,
                                                        registered_by=hospital_id,
                                                        value=fairified_value,
                                                        counter=self.counter,
                                                        dataset=dataset)
                        else:
                            raise NotImplementedError("Not implemented yet.")
                        self.records.append(new_record.to_json())
                        if len(self.records) >= BATCH_SIZE:
                            self.process_batch_of_records()
                        # to compute the percentage of missing values in features' profiles
                        if feature_id not in self.mapping_column_all_count:
                            self.mapping_column_all_count[feature_id] = 1
                        else:
                            self.mapping_column_all_count[feature_id] = self.mapping_column_all_count[feature_id] + 1
                    else:
                        # this represents the case when a column has not been converted to a Feature resource
                        # this may happen for ID column for instance, or in BUZZI many clinical columns are not described in the metadata, thus skipped here
                        # log.error(f"Skipping column {column_name} for row {index}")
                        pass
        # save the remaining tuples that have not been saved (because there were less than BATCH_SIZE tuples before the loop ends).
        if len(self.records) > 0:
            self.process_batch_of_records()
        # and save the total counts for each column (to compute the percentage of missing values in the profiles)
        all_counts = []  # a list of <"identifier": identifier, "all_counts": all_counts> instead of <identifier: all_counts>
        # log.info(self.mapping_column_all_count)
        for k in self.mapping_column_all_count:
            all_counts.append({Resource.IDENTIFIER_: k, "count_all_values": self.mapping_column_all_count[k], Resource.DATASET_: self.dataset_instance.global_identifier})
        if len(all_counts) > 0:
            self.database.insert_many_tuples(table_name=TableNames.COUNTS_FEATURES, tuples=all_counts)
        else:
            log.info(all_counts)
            log.error(f"No record counts")

    ##############################################################
    # OTHER ENTITIES
    ##############################################################

    def create_patients(self) -> None:
        log.info(f"create Patient instances in memory")
        log.info(self.execution.patient_id_column_name)
        if self.execution.patient_id_column_name not in self.data.columns:
            # no patient ID in this dataset
            log.error(f"The column {self.execution.patient_id_column_name} has been declared as the patient id but has not been found in the data. Creating automatically patient IDs.")
            pids = [i for i in range(1, len(self.data)+1)]
            self.data[self.execution.patient_id_column_name] = pids

        columns = self.data.columns
        log.info(columns)
        log.info(f"creating patients using column {self.execution.patient_id_column_name}, with index {columns.get_loc(self.execution.patient_id_column_name)}")
        for row in self.data.itertuples(index=False):
            row_patient_id = row[columns.get_loc(self.execution.patient_id_column_name)]
            if row_patient_id == "":
                # this line does not contain a patient id (the cell is empty)
                # thus, we skip it
                pass
            else:
                if row_patient_id not in self.patient_ids_mapping:
                    # the (anonymized) patient does not exist yet, we will create it
                    new_patient = Patient(identifier=NO_ID, counter=self.counter)
                    # log.info(f"create new patient {row_patient_id} with anonymized ID {new_patient.identifier.value}")
                    self.patient_ids_mapping[row_patient_id] = new_patient.identifier  # keep track of anonymized patient ids
                else:
                    # the (anonymized) patient id already exists, we take it from the mapping
                    # log.info(f"create patient {row_patient_id} with existing anonymized ID {self.patient_ids_mapping[row_patient_id]}")
                    new_patient = Patient(identifier=self.patient_ids_mapping[row_patient_id], counter=self.counter)
                self.patients.append(new_patient.to_json())

                self.time_statistics.start(dataset=self.dataset_instance.global_identifier, key=TimerKeys.CHECK_BATCH_SIZE_PATIENTS)
                if len(self.patients) >= BATCH_SIZE:
                    self.process_batch_of_patients()
                # no need to load Patient instances because they are referenced using their ID,
                # which was provided by the hospital (thus is known by the dataset)
        if len(self.patients) > 0:
            self.process_batch_of_patients()
        # finally, we also write the mapping patient ID / anonymized ID in a file - this will be ingested for subsequent runs to not renumber existing anonymized patients
        with open(self.execution.anonymized_patient_ids_filepath, "w") as data_file:
            try:
                ujson.dump(self.patient_ids_mapping, data_file)
            except Exception:
                raise ValueError(
                    f"Could not dump the {len(self.patient_ids_mapping)} JSON resources in the file located at {self.execution.anonymized_patient_ids_filepath}.")
        self.database.load_json_in_table(table_name=TableNames.PATIENT, unique_variables=[Resource.IDENTIFIER_], dataset_id=self.dataset_id)

    ##############################################################
    # UTILITIES
    ##############################################################

    def process_batch_of_patients(self) -> None:
        log.info(f"writing {len(self.patients)} patients in file")
        write_in_file(resource_list=self.patients,
                      current_working_dir=self.execution.working_dir_current,
                      table_name=TableNames.PATIENT,
                      is_feature=False,
                      dataset_id=self.dataset_id,
                      to_json=False)
        self.patients.clear()

    def process_batch_of_features(self) -> None:
        log.info(f"writing {len(self.features)} features in file")
        write_in_file(resource_list=self.features,
                      current_working_dir=self.execution.working_dir_current,
                      table_name=TableNames.FEATURE,
                      is_feature=True,
                      dataset_id=self.dataset_id,
                      to_json=False)
        self.features.clear()

    def process_batch_of_records(self) -> None:
        log.info(f"writing {len(self.records)} records in file")
        write_in_file(resource_list=self.records,
                      current_working_dir=self.execution.working_dir_current,
                      table_name=TableNames.RECORD,
                      is_feature=False,
                      dataset_id=self.dataset_id,
                      to_json=False)
        self.records.clear()

    def load_patient_id_mapping(self) -> None:
        log.info(f"Patient ID mapping filepath is {self.execution.anonymized_patient_ids_filepath}")

        # index_col is False to not add a column with line numbers
        self.patient_ids_mapping = {}
        log.debug(self.execution.anonymized_patient_ids_filepath)
        if self.execution.anonymized_patient_ids_filepath is not None:
            with open(self.execution.anonymized_patient_ids_filepath, "r") as f:
                self.patient_ids_mapping = ujson.load(f)
        log.info(f"{len(self.patient_ids_mapping)} patient IDs in the mapping file.")

    def create_ontology_resource_from_row(self, column_name: str) -> OntologyResource | None:
        rows = self.metadata[self.metadata[MetadataColumns.COLUMN_NAME].values == column_name]
        if len(rows) == 1:
            row = rows.iloc[0]
            onto_code = row.iloc[self.metadata.columns.get_loc(MetadataColumns.ONTO_CODE)]
            if len(onto_code) > 0:
                onto_system = Ontologies.get_enum_from_name(row.iloc[self.metadata.columns.get_loc(MetadataColumns.ONTO_NAME)])
                if type(onto_system) is dict and len(onto_system) > 0:
                    the_or = OntologyResource(system=onto_system, code=onto_code, label=None, quality_stats=self.quality_stats)
                    return the_or
                else:
                    log.error(
                        f"In the metadata, {MetadataColumns.ONTO_NAME} is empty or unrecognised for the column '{column_name}'.")
                    return None
            else:
                log.error(f"In the metadata, {MetadataColumns.ONTO_CODE} is empty for the column '{column_name}'.")
                return None
        elif len(rows) == 0:
            log.error(f"Did not find the column {column_name} in the metadata")
            return None
        else:
            log.error(f"Found several times the column {column_name} in the metadata")
            return None

    def fairify_value(self, column_name: str, value: Any) -> str | float | datetime | OntologyResource:
        if pd.isnull(value):
            # this is an explicit NaN value, we keep it as such (and do not convert it to None) in oder to insert it in mongodb
            return DEFAULT_NAN_VALUE
        else:
            return_value = None  # to ease logging, we also save the return value in a variable and return it at the very end
            etl_type = self.mapping_column_to_type[column_name] if column_name in self.mapping_column_to_type else DataTypes.STRING
            expected_unit = self.mapping_column_to_unit[column_name] if column_name in self.mapping_column_to_unit else None  # there was some unit specified in the metadata or extracted from the data

            # log.debug(f"ETL type is {etl_type}, expected unit is {expected_unit}")
            if etl_type == DataTypes.STRING:
                return value  # the value is already normalized, we can return it as is
            elif etl_type == DataTypes.LIST:
                return_value = value.split(",")
                return_value = [elem.strip() for elem in return_value]  # trim spaces around each str value
            elif etl_type == DataTypes.CATEGORY:
                # we look for the CC associated to that categorical value
                # we need to check that (a) the column expects this categorical value and (b) this categorical has an associated CC
                if column_name in self.mapping_column_to_categorical_value and value in self.mapping_column_to_categorical_value[column_name]:
                    return_value = self.mapping_column_to_categorical_value[column_name][value]
                else:
                    # no categorical value for that value, we return the normalized value
                    self.quality_stats.add_unknown_categorical_value(column_name=column_name, categorical_value=value)
                    return_value = value
            elif etl_type == DataTypes.API:
                # there is no pre-defined list of the possible values, instead we create a CC based on the cell value,
                # which is an ontology resource
                split_value = value.split(":")  # we consider the ontology resource is of the form <ontology name>:<code>
                ontology_name = Ontologies.normalize_name(ontology_name=split_value[0])
                ontology_code = split_value[1]  # the code will be later normalized during the CC construction
                if value not in self.mapping_apivalue_to_onto_resource:
                    onto_resource = OntologyResource(
                        system=Ontologies.get_enum_from_name(ontology_name=ontology_name), code=ontology_code, label=None,
                        quality_stats=self.quality_stats)
                    self.mapping_apivalue_to_onto_resource[value] = onto_resource
                    return_value = onto_resource
                else:
                    return_value = self.mapping_apivalue_to_onto_resource[value]
            elif etl_type == DataTypes.DATETIME or etl_type == DataTypes.DATE:
                return_value = cast_str_to_datetime(str_value=value)
            elif etl_type == DataTypes.BOOLEAN:
                value = "1" if value == "1.0" else "0" if value == "0.0" else value
                return_value = cast_str_to_boolean(str_value=value)
                if return_value is None:
                    self.quality_stats.add_unknown_boolean_value(column_name=column_name, boolean_value=value)
                    return_value = value
            elif etl_type == DataTypes.INTEGER or etl_type == DataTypes.FLOAT:
                if column_name == self.execution.patient_id_column_name:
                    # do not cast int-like string identifiers as integers because this may add too much normalization
                    return_value = str(value)
                else:
                    # this is really a numeric value that we want to cast
                    m = re.search(PATTERN_VALUE_UNIT, value)
                    if m is not None:
                        # there is a unit in the data value
                        # m.group(0) is the text itself, m.group(1) is the int/float value, m.group(2) is the unit
                        the_value = m.group(1)
                        unit = m.group(2)
                        if unit == expected_unit:
                            if etl_type == DataTypes.INTEGER:
                                return_value = cast_str_to_int(str_value=the_value)
                            elif etl_type == DataTypes.FLOAT:
                                return_value = cast_str_to_float(str_value=the_value)
                        else:
                            # the feature unit does not correspond, we return the normalized (string) value
                            self.quality_stats.add_numerical_value_with_unmatched_unit(column_name=column_name,
                                                                                            expected_unit=expected_unit,
                                                                                            current_unit=unit,
                                                                                            value=value)
                            return_value = value
                    else:
                        # this value does not contain a unit or is not of the form "value unit"
                        # thus, we cast it depending on the ETL type
                        if etl_type == DataTypes.INTEGER:
                            return_value = cast_str_to_int(str_value=value)
                        elif etl_type == DataTypes.FLOAT:
                            return_value = cast_str_to_float(str_value=value)
                        else:
                            return_value = value
            else:
                # Unhandled ETL type: this cannot happen because all ETL types have been checked during the Extract step
                return_value = value

            # in case, casting the value returned None, we set the normalized value back
            # otherwise, we keep the cast value
            return_value = value if return_value is None else return_value
            # count how many fairified values do not (still) match the expected ETL type
            if ((etl_type == DataTypes.BOOLEAN and not isinstance(return_value, bool))
                    or (etl_type == DataTypes.FLOAT and not isinstance(return_value, float))
                    or (etl_type == DataTypes.INTEGER and not isinstance(return_value, int))
                    or (etl_type == DataTypes.DATE and not isinstance(return_value, datetime))
                    or (etl_type == DataTypes.DATETIME and not isinstance(return_value, datetime))
                    or (etl_type == DataTypes.STRING and not isinstance(return_value, str))
                    or (etl_type == DataTypes.CATEGORY and not isinstance(return_value, OntologyResource))):
                self.quality_stats.add_column_with_unmatched_typeof_etl_types(column_name=column_name,
                                                                              typeof_type=type(return_value).__name__,
                                                                              etl_type=etl_type)

            # we use type(..).__name__ to get the class name, e.g., "str" or "bool", instead of "<class 'float'>"
            # log.info(f"Column '{column_name}': fairify {type(value).__name__} value '{value}' (unit: {expected_unit}) into {type(return_value).__name__}: {return_value}")
            return return_value

    def anonymize_value(self, column_name: str, fairified_value: Any) -> tuple:
        """

        :param column_name:
        :type column_name:
        :param fairified_value:
        :type fairified_value:
        :return: two values: either (the anonymized value, True), or (the fairified value, False). This is to know whether we should create another Record with the anonymized value
        :rtype:
        """
        etl_type = self.mapping_column_to_type[column_name] if column_name in self.mapping_column_to_type else DataTypes.STRING
        visibility = self.mapping_column_to_visibility[column_name] if column_name in self.mapping_column_to_visibility else Visibility.PRIVATE
        if etl_type == DataTypes.DATETIME:
            if visibility == Visibility.ANONYMIZED:
                # anonymize the date and the time
                # since a datetime object always contains day+month+year (in any order), we cannot get rid of the day
                # however, we can set it to 01
                # similarly for hour:minute:second, we set it to 0
                anonymized_value = fairified_value.replace(day=1)
                anonymized_value = anonymized_value.replace(hour=0)
                anonymized_value = anonymized_value.replace(minute=0)
                anonymized_value = anonymized_value.replace(second=0)
                return anonymized_value, True
            else:
                # no need to anonymize the datetime
                return fairified_value, False
        elif etl_type == DataTypes.DATE:
            if visibility == Visibility.ANONYMIZED:
                # anonymize the date
                anonymized_value = fairified_value.replace(day=1)
                return anonymized_value, True
            else:
                # no need to anonymize the date
                return fairified_value, False
        else:
            return fairified_value, False



import dataclasses
import json

from constants.methods import factory


@dataclasses.dataclass()
class DatasetProfile:
    description: str
    theme: str
    filetype: str
    size: float  # in Mb
    nb_tuples: int
    completeness: int
    uniqueness: float

    def to_json(self):
        return {"ds_profile": "the_ds_profile"}
        # return dataclasses.asdict(self, dict_factory=factory)

    def __str__(self):
        return json.dumps(self.to_json())



import logging

import headfake.fieldset
import urllib3

logging.basicConfig(
    level=logging.DEBUG,
    # format='%(asctime)s, %(levelname)-5s [%(module)s:%(funcName)s:%(lineno)d] %(message)s',
    format='[%(module)s:%(funcName)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    handlers=[
        logging.StreamHandler()
        # we will add a FileHandler in the Execution class when we know the db name (user input)
    ]
)
log = logging.getLogger()
# do not allow PyMongo to print everything,
# only important messages (warning, error and fatal) wil be shown
logging.getLogger("pymongo").setLevel(logging.WARNING)
logging.getLogger("matplotlib").setLevel(logging.WARNING)
logging.getLogger("PIL").setLevel(logging.WARNING)
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.ERROR)
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logging.getLogger("faker").setLevel(logging.ERROR)
logging.getLogger("factory").setLevel(logging.ERROR)
logging.getLogger("headfake").setLevel(logging.ERROR)
logging.getLogger("faker").setLevel(logging.ERROR)
logging.getLogger("fieldset").setLevel(logging.ERROR)




import json
import os
import re
import time
import unittest

import pandas as pd
import pytest

from constants.defaults import DEFAULT_ONTOLOGY_RESOURCE_LABEL, NO_ID
from constants.structure import TEST_DB_NAME, DOCKER_FOLDER_TEST
from database.Counter import Counter
from database.Database import Database
from entities.Dataset import Dataset
from database.Execution import Execution
from entities.Feature import Feature
from entities.Hospital import Hospital
from entities.OntologyResource import OntologyResource
from entities.Record import Record
from entities.Resource import Resource
from enums.DataTypes import DataTypes
from enums.HospitalNames import HospitalNames
from enums.MetadataColumns import MetadataColumns
from enums.Ontologies import Ontologies
from enums.ParameterKeys import ParameterKeys
from enums.Profile import Profile
from enums.TableNames import TableNames
from enums.TheTestFiles import TheTestFiles
from enums.Visibility import Visibility
from etl.Transform import Transform
from main_statistics.QualityStatistics import QualityStatistics
from utils.cast_utils import cast_str_to_datetime
from utils.file_utils import get_json_resource_file, from_json_line_to_json_str, clear_file
from utils.file_utils import read_tabular_file_as_string
from utils.setup_logger import log
from utils.test_utils import set_env_variables_from_dict, get_feature_by_text, \
    get_field_value_for_patient, get_records_for_patient


# personalized setup called at the beginning of each test
def my_setup(hospital_name: str, profile: str, extracted_metadata_path: str, extracted_data_paths: str,
             extracted_column_to_categorical_path: str,
             extracted_column_type_path: str,
             extracted_column_unit_path: str,
             extracted_domain_path: str,
             db_drop: str = "True") -> Transform:
    args = {
        ParameterKeys.DB_NAME: TEST_DB_NAME,
        ParameterKeys.DB_DROP: db_drop,
        ParameterKeys.HOSPITAL_NAME: hospital_name
        # no need to set the metadata and data filepaths as we get already the loaded data and metadata as arguments
    }
    set_env_variables_from_dict(env_vars=args)
    TestTransform.execution.internals_set_up()
    TestTransform.execution.file_set_up(setup_files=False)  # no need to set up the files, we get data and metadata as input
    TestTransform.execution.db_drop = True
    database = Database(execution=TestTransform.execution)
    # I load:
    # - the data and metadata from two CSV files that I obtained by running the Extract step
    # - and mapped_values as a JSON file that I obtained from the same Extract object
    metadata = read_tabular_file_as_string(os.path.join(DOCKER_FOLDER_TEST, extracted_metadata_path))
    data_path = os.path.join(DOCKER_FOLDER_TEST, extracted_data_paths)
    data = read_tabular_file_as_string(data_path)
    # for data only, we need to do a second pass on the data to replace explicit NaN values by np.nan
    # something Extract is doing, but here we need to do it because we simply read the expected CSV file,
    # thus NaN values are read as strings and need to be transformed
    for column in data:
        data.loc[:, column] = data[column].apply(lambda x: MetadataColumns.normalize_value(column_value=x))

    with open(os.path.join(DOCKER_FOLDER_TEST, extracted_column_to_categorical_path), "r") as f:
        mapping_column_to_categorical_value = json.load(f)
    with open(os.path.join(DOCKER_FOLDER_TEST, extracted_column_unit_path), "r") as f:
        column_to_unit = json.load(f)
    with open(os.path.join(DOCKER_FOLDER_TEST, extracted_domain_path), "r") as f:
        mapping_column_to_domain = json.load(f)
    with open(os.path.join(DOCKER_FOLDER_TEST, extracted_column_type_path), "r") as f:
        mapping_column_to_type = json.load(f)

    dataset_instance = Dataset(identifier=NO_ID, database=database, docker_path=data_path, version_notes=None, license=None, counter=Counter())
    transform = Transform(database=database, execution=TestTransform.execution, data=data, metadata=metadata,
                          profile=profile, dataset_id=get_dataset_id_from_profile(profile),
                          mapping_column_to_categorical_value=mapping_column_to_categorical_value,
                          mapping_column_to_unit=column_to_unit,
                          mapping_column_to_domain=mapping_column_to_domain,
                          mapping_column_to_type=mapping_column_to_type,
                          load_patients=True,
                          quality_stats=QualityStatistics(record_stats=False),
                          dataset_key=dataset_instance)

    # create a hospital instance, to be able to create records with references to the hospital
    transform.database.insert_one_tuple(table_name=TableNames.HOSPITAL, one_tuple=Hospital(identifier=NO_ID, name=HospitalNames.TEST_H1, counter=Counter()).to_json())
    counter = Counter()
    counter.set_with_database(database=transform.database)
    return transform


def get_dataset_id_from_profile(profile: str):
    if profile == Profile.PHENOTYPIC:
        return 1
    elif profile == Profile.CLINICAL:
        return 2
    elif profile == Profile.DIAGNOSIS:
        return 3
    else:
        return 0


@pytest.fixture(autouse=True)
def run_before_and_after_tests(tmpdir):
    """Fixture to execute asserts before and after a test is run"""
    # Setup: fill with any logic you want

    yield  # this is where the testing happens

    # Teardown : fill with any logic you want
    write_empty_pid_files()


def delete_pid_file():
    pid_path = os.path.join(DOCKER_FOLDER_TEST, TheTestFiles.PIDS_PATH)
    if os.path.exists(pid_path):
        os.remove(pid_path)


def create_pid_file():
    with open(os.path.join(DOCKER_FOLDER_TEST, TheTestFiles.PIDS_PATH), "w") as f:
        the_pids = json.dumps({
            "999999999": 999,
            "999999998": 998,
            "999999997": 997,
            "999999996": 996,
            "999999995": 995,
            "999999994": 994,
            "999999993": 993,
            "999999992": 992,
            "999999991": 991,
            "999999990": 990
        })
        f.write(the_pids)


def write_empty_pid_files():
    with open(os.path.join(DOCKER_FOLDER_TEST, TheTestFiles.PIDS_PATH), "w") as f:
        f.write(json.dumps({}))


def get_transform_features(profile):
    with open(get_json_resource_file(current_working_dir=TestTransform.execution.working_dir_current, dataset_id=get_dataset_id_from_profile(profile), table_name=TableNames.FEATURE)) as f:
        # return json.load(f)  # this is a JSON-line file, we need to append brackets and commas back before parsing it
        return json.loads(from_json_line_to_json_str(f))


def get_transform_records(profile):
    with open(get_json_resource_file(current_working_dir=TestTransform.execution.working_dir_current, dataset_id=get_dataset_id_from_profile(profile), table_name=TableNames.RECORD)) as f:
        # return json.load(f)  # this is a JSON-line file, we need to append brackets and commas back before parsing it
        return json.loads(from_json_line_to_json_str(f))


def get_transform_patients(dataset_id):
    with open(get_json_resource_file(current_working_dir=TestTransform.execution.working_dir_current, dataset_id=dataset_id, table_name=TableNames.PATIENT)) as f:
        # return json.load(f)  # this is a JSON-line file, we need to append brackets and commas back before parsing it
        return json.loads(from_json_line_to_json_str(f))


class TestTransform(unittest.TestCase):
    execution = Execution()

    def test_set_resource_counter_id(self):
        # when there is nothing in the database, the counter should be 0
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.PHENOTYPIC,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_PHENOTYPIC_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_PHENOTYPIC_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_PHENOTYPIC_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_PHENOTYPIC_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_PHENOTYPIC_DOMAIN_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_PHENOTYPIC_TYPE_PATH)
        counter = Counter()
        counter.set_with_database(database=transform.database)

        assert transform.counter.resource_id == 0

        # when some tables already contain resources, it should not be 0
        # I manually insert some resources in the database
        database = Database(execution=TestTransform.execution)
        my_tuples = [
            {"identifier": 1},
            {"identifier": 2},
            {"identifier": 7},
            {"identifier": 3},
            {"identifier": 124},
            {"identifier": 9},
            {"identifier": 123},
            {"identifier": 6}
        ]
        database.insert_many_tuples(table_name=TableNames.FEATURE, tuples=my_tuples)
        time.sleep(2)
        counter.set_with_database(database=transform.database)
        assert counter.resource_id == 125

    def test_phenotypic_data(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.PHENOTYPIC,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_PHENOTYPIC_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_PHENOTYPIC_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_PHENOTYPIC_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_PHENOTYPIC_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_PHENOTYPIC_DOMAIN_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_PHENOTYPIC_TYPE_PATH)
        transform.create_patients()
        transform.counter.set_with_database(database=transform.database)
        transform.create_features()
        transform.counter.set_with_database(database=transform.database)
        transform.create_records()

        # CHECK FEATURES
        features = get_transform_features(profile=Profile.PHENOTYPIC)
        assert len(features) == 4 - 1  # id does not count as a PhenFeature
        # assert the fourth and first PhenFeature instances:
        # lab_feature_a has one associated code
        # lab_feature_b has no associated code
        # because they may not be in the same order as in the metadata file, we get them based on their text
        # (which contains at least the column name, and maybe a description)
        # PhenFeature about sex
        lab_feature_a = get_feature_by_text(features, "sex")
        assert len(lab_feature_a) == 11  # inherited fields (identifier, entity_type, timestamp), proper fields (name, ontology_resource, data_type, unit, visibility, datasets, description, domain)
        assert Resource.IDENTIFIER_ in lab_feature_a
        assert lab_feature_a[Feature.NAME_] == "sex"
        assert lab_feature_a[Feature.ONTO_] == {
            OntologyResource.SYSTEM_: Ontologies.SNOMEDCT["url"],
            OntologyResource.CODE_: "123:789",
            OntologyResource.LABEL_: f"{DEFAULT_ONTOLOGY_RESOURCE_LABEL}:{DEFAULT_ONTOLOGY_RESOURCE_LABEL}"
            # the two codes do not exist for eal in SNOMED, thus using the empty label
        }
        assert lab_feature_a[Feature.DT_] == DataTypes.CATEGORY
        assert Feature.UNIT_ not in lab_feature_a
        assert lab_feature_a[Feature.ENTITY_TYPE_] == f"{Profile.PHENOTYPIC}{TableNames.FEATURE}"
        assert lab_feature_a[Feature.DOMAIN_] is not None
        assert "accepted_values" in lab_feature_a[Feature.DOMAIN_]
        assert len(lab_feature_a[Feature.DOMAIN_]["accepted_values"]) == 2
        assert "m" in lab_feature_a[Feature.DOMAIN_]["accepted_values"]
        assert "f" in lab_feature_a[Feature.DOMAIN_]["accepted_values"]

        # PhenFeature about date_of_birth
        lab_feature_b = get_feature_by_text(features, "date_of_birth")
        assert len(lab_feature_b) == 8  # no ontology_resource, no unit, no domain
        assert Resource.IDENTIFIER_ in lab_feature_b
        assert Resource.TIMESTAMP_ in lab_feature_b
        assert lab_feature_b[Feature.NAME_] == "date_of_birth"
        assert Feature.ONTO_ not in lab_feature_b
        assert lab_feature_b[Feature.DT_] == DataTypes.DATETIME
        assert Feature.UNIT_ not in lab_feature_b
        assert lab_feature_b[Feature.ENTITY_TYPE_] == f"{Profile.PHENOTYPIC}{TableNames.FEATURE}"
        assert lab_feature_b[Feature.DESCR_] == "The date of birth"
        assert lab_feature_b[Feature.VISIBILITY_] == Visibility.ANONYMIZED

        # check that there are no duplicates in PhenFeature instances
        # for this, we get the set of their names (in the field "text")
        lab_features_names_list = []
        for lab_feature in features:
            if lab_feature is not None:
                lab_features_names_list.append(lab_feature[Feature.NAME_])
        lab_features_names_set = set(lab_features_names_list)
        assert len(lab_features_names_list) == len(lab_features_names_set)

        # CHECK RECORDS
        records = get_transform_records(profile=Profile.PHENOTYPIC)
        log.info(records)
        assert len(records) == 18  # in total, 18 PhenRecord instances are created, between 2 and 5 per Patient, the only explicit NaN value is indeed created
        # assert that PhenRecord instances have been correctly created for a given data row
        # we take the seventh row
        patient_id = transform.patient_ids_mapping["999999994"]
        assert patient_id == 6  # patient anonymized IDs start at h1:1, because no hospital has been created beforehand.
        phen_records_patient = get_records_for_patient(records=records, patient_id=patient_id)
        log.info(phen_records_patient)
        female = OntologyResource(system=Ontologies.SNOMEDCT, code="248152002", label=None, quality_stats=None)
        assert phen_records_patient[0][Record.VALUE_] == female.to_json()  # the value as been replaced by its ontology code (sex is a categorical value)
        assert phen_records_patient[1][Record.VALUE_] == "black"
        assert phen_records_patient[2][Record.VALUE_] == {"$date": "2021-12-01T00:00:00Z"}  # the value as been converted to a MongoDB-style datetime and anonymized (remove day and time)
        pattern_date = re.compile("[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2,3}Z")
        assert pattern_date.match(phen_records_patient[0][Resource.TIMESTAMP_]["$date"])  # check the date is in datetime (CEST) form

        # we also check that string are normalized (no caps, etc.)
        assert transform.patient_ids_mapping["999999996"] == 4  # 0 Hospital before patients
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999999"], column_name="ethnicity") == "white"
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999998"], column_name="ethnicity") == "white"
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999997"], column_name="ethnicity") == "white"
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999996"], column_name="ethnicity") == "caucasian"
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999995"], column_name="ethnicity") is None  # no value so None
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999994"], column_name="ethnicity") == "black"
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999993"], column_name="ethnicity") == "caucasian"
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999992"], column_name="ethnicity") == "italian"
        assert pd.isnull(get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999991"], column_name="ethnicity"))  # explicit NaN value
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999990"], column_name="ethnicity") is None  # no value so None

    def test_clinical_data(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.CLINICAL,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_CLINICAL_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_CLINICAL_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_CLINICAL_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_CLINICAL_DOMAIN_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_CLINICAL_TYPE_PATH)
        transform.load_patient_id_mapping()
        transform.create_patients()
        transform.counter.set_with_database(database=transform.database)
        transform.create_features()
        transform.counter.set_with_database(database=transform.database)
        transform.create_records()

        # CHECK FEATURES
        # we cannot use transform.features, because the array is cleared after all features are saved
        # we need to read from the JSON file written during the Transform step
        features = get_transform_features(profile=Profile.CLINICAL)
        log.info(features)
        assert len(features) == 6 - 2  # sid and id do not count as SamFeatures
        # assert the third and fourth SamFeature instances:
        # lab_feature_a has one associated code
        # lab_feature_b has no associated code
        # because they may not be in the same order as in the metadata file, we get them based on their text
        # (which contains at least the column name, and maybe a description)
        # SamFeature about molecule_a
        lab_feature_a = get_feature_by_text(features, "molecule_a")
        assert len(lab_feature_a) == 11  # inherited fields (identifier, entity_type, timestamp), proper fields (name, ontology_resource, data_type, unit, visibility, datasets, description, domain)
        assert Resource.IDENTIFIER_ in lab_feature_a
        assert lab_feature_a[Feature.NAME_] == "molecule_a"
        assert lab_feature_a[Feature.ONTO_] == {
            OntologyResource.SYSTEM_: Ontologies.LOINC["url"],
            OntologyResource.CODE_: "1234",
            OntologyResource.LABEL_: DEFAULT_ONTOLOGY_RESOURCE_LABEL
            # this resource does not exist for real in LOINC, thus display is empty
        }
        assert lab_feature_a[Feature.DT_] == DataTypes.FLOAT
        assert lab_feature_a[Feature.UNIT_] == "mg/L"
        assert lab_feature_a[Feature.VISIBILITY_] == Visibility.PUBLIC
        assert lab_feature_a[Feature.ENTITY_TYPE_] == f"{Profile.CLINICAL}{TableNames.FEATURE}"
        assert lab_feature_a[Feature.DESCR_] == "The molecule Alpha"
        assert lab_feature_a[Feature.DOMAIN_] == {"min": 0}
        # timestamp is not tested

        # LabFeature about molecule_b
        lab_feature_b = get_feature_by_text(features, "molecule_b")
        assert len(lab_feature_b) == 9
        assert Resource.IDENTIFIER_ in lab_feature_b
        assert lab_feature_b[Feature.NAME_] == "molecule_b"
        assert Feature.ONTO_ not in lab_feature_b
        assert lab_feature_b[Feature.DT_] == DataTypes.INTEGER
        assert lab_feature_b[Feature.UNIT_] == "g"  # unit is gram
        assert lab_feature_b[Feature.ENTITY_TYPE_] == f"{Profile.CLINICAL}{TableNames.FEATURE}"
        assert lab_feature_b[Feature.DESCR_] == "The molecule Beta"
        assert Feature.DOMAIN_ not in lab_feature_b

        # check that there are no duplicates in SamFeature instances
        # for this, we get the set of their names (in the field "text")
        lab_features_names_list = []
        for lab_feature in features:
            if lab_feature is not None:
                lab_features_names_list.append(lab_feature[Feature.NAME_])
        lab_features_names_set = set(lab_features_names_list)
        assert len(lab_features_names_list) == len(lab_features_names_set)

        # CHECK RECORDS
        records = get_transform_records(profile=Profile.CLINICAL)
        log.info(records)
        assert len(records) == 23  # in total, 16 ClinicalRecord instances are created, between 2 and 5 per Patient

        # assert that ClinicalRecord instances have been correctly created for a given data row
        # we take the seventh row
        patient_id = transform.patient_ids_mapping["999999994"]
        log.info(transform.patient_ids_mapping)
        assert patient_id == 6
        records_patient = get_records_for_patient(records=records, patient_id=patient_id)
        assert len(records_patient) == 2
        assert records_patient[0][Record.VALUE_] == -0.003  # the value as been converted to a float
        assert records_patient[0][Record.SUBJECT_] == patient_id
        assert records_patient[0][Record.REG_BY_] == 1  # Hospital:1
        assert records_patient[0][Record.INSTANTIATES_] == 12  # LabRecord 12 is about molecule_a (there are Hospital 1 and Dataset 2, and patients with normal anonymized pids)
        assert records_patient[0][Feature.ENTITY_TYPE_] == f"{Profile.CLINICAL}{TableNames.RECORD}"  #
        pattern_date = re.compile("[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2,3}Z")
        assert pattern_date.match(records_patient[0][Resource.TIMESTAMP_]["$date"])  # check the date is in datetime (CEST) form

        # check that all values are cast to the expected type
        assert records_patient[0][Record.VALUE_] == -0.003  # the value as been converted to an integer
        assert records_patient[1][Record.VALUE_] is False  # the value as been converted to a boolean

        # we also check that conversions str->int/float and category->bool worked
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999999"], column_name="molecule_b") == 100  # this has been cast as int because it matches the expected unit
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999998"], column_name="molecule_b") == 111  # this has been cast as int because it matches the expected unit
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999997"], column_name="molecule_b") == "231 grams"  # this has not been converted as this does not match the expected unit
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999996"], column_name="molecule_b") == 21  # this has been cast as int because it matches the expected unit
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999995"], column_name="molecule_b") == 100  # this has been cast as int even though there was no unit
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999994"], column_name="molecule_b") is None  # no value, thus no Record (thus None)
        assert pd.isnull(get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999993"], column_name="molecule_b"))  # null value converted to NaN
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999992"], column_name="molecule_b") == "116 kg"  # this has not been converted as this does not match the expected unit
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999991"], column_name="molecule_b") is None  # no value, thus no Record
        assert get_field_value_for_patient(records=records, features=features, patient_id=transform.patient_ids_mapping["999999990"], column_name="molecule_b") is None  # no value, thus no Record
    # TODO Nelly: check there are no duplicates for SamFeature instances

    def test_create_patients_without_pid(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.PHENOTYPIC,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_PHENOTYPIC_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_PHENOTYPIC_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_PHENOTYPIC_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_PHENOTYPIC_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_PHENOTYPIC_DOMAIN_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_PHENOTYPIC_TYPE_PATH)
        # before creating the patients, we need to erase the content of the patient file
        # that has been written by test_create_patients_without_pid
        # otherwise, it would count 20 patients instead of 10 (2 times)
        # we do not need to do this trick in the normal execution because each time a new working directory is created
        # thus, there are no overlapping files
        clear_file(current_working_dir=transform.execution.working_dir_current,
                   dataset_id=get_dataset_id_from_profile(profile=Profile.PHENOTYPIC),
                   table_name=TableNames.PATIENT)

        # this creates Patient resources (based on the data file) and insert them in a (JSON) temporary file
        transform.create_patients()

        patients = get_transform_patients(dataset_id=get_dataset_id_from_profile(profile=Profile.PHENOTYPIC))
        assert len(patients) == 10
        # we cannot simply order by identifier value because they are strings, not int
        # thus will need a bit more of processing to sort by the integer represented within the string
        sorted_patients = sorted(patients, key=lambda d: d[Resource.IDENTIFIER_])
        for i in range(0, len(sorted_patients)):
            # patients have their own anonymized ids
            assert sorted_patients[i][Resource.IDENTIFIER_] == i + 1

        # get back to the original file
        with open(self.execution.anonymized_patient_ids_filepath, "w") as f:
            f.write("{}")

    def test_create_patients_with_pid(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.PHENOTYPIC,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_PHENOTYPIC_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_PHENOTYPIC_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_PHENOTYPIC_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_PHENOTYPIC_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_PHENOTYPIC_DOMAIN_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_PHENOTYPIC_TYPE_PATH)
        # before creating the patients, we need to erase the content of the patient file
        # that has been written by test_create_patients_without_pid
        # otherwise, it would count 20 patients instead of 10 (2 times)
        # we do not need to do this trick in the normal execution because each time a new working directory is created
        # thus, there are no overlapping files
        clear_file(current_working_dir=transform.execution.working_dir_current,
                   dataset_id=get_dataset_id_from_profile(profile=Profile.PHENOTYPIC),
                   table_name=TableNames.PATIENT)

        # this creates Patient resources (based on the data file) and insert them in a (JSON) temporary file
        transform.load_patient_id_mapping()
        transform.create_patients()

        patients = get_transform_patients(dataset_id=get_dataset_id_from_profile(profile=Profile.PHENOTYPIC))
        assert len(patients) == 10
        # we cannot simply order by identifier value because they are strings, not int
        # thus will need a bit more of processing to sort by the integer represented within the string
        sorted_patients = sorted(patients, key=lambda d: d[Resource.IDENTIFIER_])
        # sorted_patients = sorted(transform.patients)
        log.info(patients)
        log.info(sorted_patients)
        for i in range(0, len(sorted_patients)):
            # patients have their own anonymized ids
            assert sorted_patients[i][Resource.IDENTIFIER_] == i + 1

    def test_create_ontology_resource_from_row(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.CLINICAL,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_CLINICAL_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_CLINICAL_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_CLINICAL_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_CLINICAL_DOMAIN_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_CLINICAL_TYPE_PATH)
        # no associated ontology code
        onto_resource = transform.create_ontology_resource_from_row(column_name="molecule_b")
        assert onto_resource is None

        # one associated ontology code
        onto_resource = transform.create_ontology_resource_from_row(column_name="molecule_a")
        assert onto_resource.to_json() == OntologyResource(system=Ontologies.LOINC, code="1234", label=None, quality_stats=None).to_json()

    def test_ontology_resource(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.CLINICAL,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_CLINICAL_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_CLINICAL_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_CLINICAL_UNITS_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_CLINICAL_TYPE_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_CLINICAL_DOMAIN_PATH)
        # no associated ontology code (patient id line)
        first_row = transform.metadata.iloc[0]
        onto_resource = OntologyResource(system=Ontologies.get_enum_from_name(first_row[MetadataColumns.ONTO_NAME]),
                                         code=first_row[MetadataColumns.ONTO_CODE],
                                         label=None, quality_stats=None)
        assert onto_resource is not None
        assert onto_resource.system == {}
        assert onto_resource.code == ""
        assert onto_resource.label is None

        # one associated ontology code (molecule_g line)
        fourth = transform.metadata.iloc[4]
        onto_resource = OntologyResource(system=Ontologies.get_enum_from_name(fourth[MetadataColumns.ONTO_NAME]),
                                         code=fourth[MetadataColumns.ONTO_CODE],
                                         label=None, quality_stats=None)
        assert onto_resource is not None
        assert onto_resource.system == Ontologies.SNOMEDCT["url"]
        assert onto_resource.code == "421416008"
        assert onto_resource.label == "Gamma"

    def test_fairify_phen_value(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.PHENOTYPIC,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_PHENOTYPIC_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_PHENOTYPIC_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_PHENOTYPIC_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_PHENOTYPIC_UNITS_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_PHENOTYPIC_TYPE_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_PHENOTYPIC_DOMAIN_PATH)

        assert transform.fairify_value(column_name="id", value=transform.data.iloc[0][0]) == "999999999"
        onto_resource = OntologyResource(system=Ontologies.SNOMEDCT, code="248152002", label=None, quality_stats=None)
        fairified_value = transform.fairify_value(column_name="sex", value=transform.data.iloc[0][1])
        assert fairified_value == onto_resource.to_json()
        assert transform.fairify_value(column_name="ethnicity", value=transform.data.iloc[0][2]) == "white"
        assert transform.fairify_value(column_name="date_of_birth", value=transform.data.iloc[0][3]) == ""  # no date here, we verify this returns empty string
        assert transform.fairify_value(column_name="date_of_birth", value=transform.data.iloc[5][3]) == cast_str_to_datetime(str_value="2021-12-22 11:58:38.881")

    def test_fairify_clin_value(self):
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.CLINICAL,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_CLINICAL_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_CLINICAL_COL_CAT_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_CLINICAL_TYPE_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_CLINICAL_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_CLINICAL_DOMAIN_PATH)

        assert transform.fairify_value(column_name="sid", value=transform.data.iloc[0][0]) == "s1"
        assert transform.fairify_value(column_name="id", value=transform.data.iloc[0][1]) == "999999999"
        assert transform.fairify_value(column_name="molecule_a", value=transform.data.iloc[0][2]) == 0.001
        assert transform.fairify_value(column_name="molecule_b", value=transform.data.iloc[0][3]) == 100
        assert transform.fairify_value(column_name="molecule_g", value=transform.data.iloc[0][4]) is True
        assert transform.fairify_value(column_name="molecule_g", value=transform.data.iloc[4][4]) is False
        assert pd.isnull(transform.fairify_value(column_name="molecule_g", value=transform.data.iloc[6][4]))
        assert pd.isnull(transform.fairify_value(column_name="molecule_g", value=transform.data.iloc[7][4]))

    def test_patient_id_mapping(self):
        # test with no patient ID mapping
        # the mapping filename HAS TO BE "anonymized_patient_ids.json"
        delete_pid_file()
        extract = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.CLINICAL,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_CLINICAL_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_CLINICAL_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_CLINICAL_UNITS_PATH,
                           extracted_domain_path=TheTestFiles.EXTR_CLINICAL_DOMAIN_PATH,
                           extracted_column_type_path=TheTestFiles.EXTR_CLINICAL_TYPE_PATH)
        extract.load_patient_id_mapping()

        # when the file is empty, the Execution should write an empty list into it
        assert os.stat(extract.execution.anonymized_patient_ids_filepath).st_size > 0
        assert extract.patient_ids_mapping == {}

        # test with non-empty patient ID mapping
        create_pid_file()
        transform = my_setup(hospital_name=HospitalNames.TEST_H1, profile=Profile.CLINICAL,
                             extracted_metadata_path=TheTestFiles.EXTR_METADATA_CLINICAL_PATH,
                             extracted_data_paths=TheTestFiles.EXTR_CLINICAL_DATA_PATH,
                             extracted_column_to_categorical_path=TheTestFiles.EXTR_CLINICAL_COL_CAT_PATH,
                             extracted_column_unit_path=TheTestFiles.EXTR_CLINICAL_UNITS_PATH,
                             extracted_domain_path=TheTestFiles.EXTR_CLINICAL_DOMAIN_PATH,
                             extracted_column_type_path=TheTestFiles.EXTR_CLINICAL_TYPE_PATH,
                             db_drop="False")
        transform.load_patient_id_mapping()

        # when the file is not empty, all mappings should be loaded in Extract
        assert os.stat(transform.execution.anonymized_patient_ids_filepath).st_size > 0
        with open(os.path.join(DOCKER_FOLDER_TEST, TheTestFiles.PIDS_PATH), "r") as f:
            expected_dict = json.load(f)
        assert expected_dict == transform.patient_ids_mapping
        delete_pid_file()  # get back to the original setting where no PID is provided



import dataclasses

from entities.Record import Record
from enums.Profile import Profile
from enums.TableNames import TableNames


@dataclasses.dataclass
class PhenotypicRecord(Record):
    entity_type: str = f"{Profile.PHENOTYPIC}{TableNames.RECORD}"



from typing import Any

import numpy as np
import pandas as pd

from enums.EnumAsClass import EnumAsClass
from utils.setup_logger import log


class Visibility(EnumAsClass):
    PUBLIC = "PUBLIC"
    ANONYMIZED = "ANONYMIZED"
    PRIVATE = "PRIVATE"

    METADATA_NB_UNRECOGNIZED_VISIBILITY = 0

    @classmethod
    def get_enum_from_name(cls, visibility_str: str):
        for existing_visibility in Visibility.values():
            if existing_visibility == visibility_str:
                return existing_visibility  # return the visibility enum
        # at the end of the loop, no enum value could match the given visibility
        # thus we need to raise an error
        raise ValueError(f"The given visibility value ({visibility_str}) does not correspond to any known visibility.")

    @classmethod
    def normalize(cls, visibility: str) -> str:
        if visibility == "":
            return ""
        else:
            visibility_u = visibility.upper()
            if visibility_u not in Visibility.values():
                log.error(f"{visibility_u} is not a recognized visibility; we will use 'PRIVATE' visibility by default.")
                Visibility.METADATA_NB_UNRECOGNIZED_VISIBILITY += 1
                return Visibility.PRIVATE
            else:
                return Visibility.get_enum_from_name(visibility)



from __future__ import annotations
from datetime import datetime
from typing import Any

from enums.EnumAsClass import EnumAsClass

THE_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class Operators(EnumAsClass):
    @classmethod
    def match(cls, field: str | None, value: Any, is_regex: bool) -> dict:
        if is_regex:
            # this is a match with a regex (in value)
            return {
                "$match": {
                    field: {
                        # the value (the regex) should not contain the / delimiters as in /^[0-9]+$/,
                        # but only the regex, as in "^[0-9]+$"
                        "$regex": value
                    }
                }
            }
        else:
            # this is a match with a "hard-coded" value (in value) or a complex operator
            if field is None:
                # the value is a complex operator, e.g., and/or
                return {
                    "$match": value
                }
            else:
                # this is a match with a "hard-coded" value (in value)
                return {
                    "$match": {
                        field: value
                    }
                }

    @classmethod
    def or_operator(cls, list_of_conditions: list[dict]) -> dict:
        # list_of_conditions is a list where each element is a dict-like condition,
        # containing the field and the condition to apply
        # e.g., [{"value": 3}, {"value": {"$type": "bool"}}]
        return {
            "$or": list_of_conditions
        }

    @classmethod
    def and_operator(cls, list_of_conditions: list[dict]) -> dict:
        # list_of_conditions is a list where each element is a dict-like condition,
        # containing the field and the condition to apply
        # e.g., [{"value": 3}, {"value": {"$type": "bool"}}]
        return {
            "$and": list_of_conditions
        }

    @classmethod
    def project(cls, field: str | list | dict | None, projected_value: str | dict | None) -> dict:
        # this is the SQL SELECT operator
        # where each field that is wanted in the result has value 1 and unwanted fields have value 0
        if type(projected_value) is str:
            # in this case, we want to keep a certain field
            # and choose what should be the value of that field (in case of composed fields)
            if type(field) is dict:
                # if the composed field may be empty, we use an alternative field ("$ifNull", within the field)
                return {
                    "$project": {
                        projected_value: field,
                        "_id": 0
                    }
                }
            else:
                # else, we simply return the projection
                return {
                    "$project": {
                        projected_value: "$" + field,
                        "_id": 0
                    }
                }
        elif type(projected_value) is dict:
            # in case we give a complex projection, e.g., with $split
            return {
                "$project": projected_value
            }
        else:
            # in that case, we only want to keep a certain field
            if isinstance(field, list):
                # we want to keep several fields
                fields = {f: 1 for f in field}
                fields["_id"] = 0
                return {
                    "$project": fields
                }
            elif isinstance(field, dict):
                # we want to compute a novel value while not keeping the old ones
                return {
                    "$project": field
                }
            else:
                # we want to keep a single field
                return {
                    "$project": {
                        field: 1,
                        "_id": 0
                    }
                }

    @classmethod
    def lookup(cls, join_table_name: str, foreign_field: str|None, local_field: str|None, lookup_field_name: str, let: dict|None,
               pipeline: list|None) -> dict:
        # this is a SQL join between two tables
        # from is the "second" table of the join
        # localField is the field of the "second" table to join
        # foreignField is the field of the "first" table to join
        # as is the name of the (new array) field added containing either the joined resource (of the second table) or an empty array if no join could be made for the tuple
        the_lookup = {
            "$lookup": {
                "from": join_table_name,
                "as": lookup_field_name,
            }
        }

        if local_field is not None:
            the_lookup["$lookup"]["localField"] = local_field
        if foreign_field is not None:
            the_lookup["$lookup"]["foreignField"] = foreign_field
        if let is not None:
            the_lookup["$lookup"]["let"] = let
        if pipeline is not None:
            the_lookup["$lookup"]["pipeline"] = pipeline
        return the_lookup

    @classmethod
    def cartesian_product(cls, join_table_name: str, lookup_field_name: str, filter_dict: dict) -> dict:
        # if len(filter_dict) == 0:
        #     pipeline = [{"$project": {field_b: 1, "_id": 0}}]
        # else:
        #     pipeline = [filter_dict, {"$project": {field_b: 1, "_id": 0}}]
        return {
            "$lookup": {
                "from": join_table_name,
                "pipeline": [filter_dict] if len(filter_dict) > 0 else [],
                "as": lookup_field_name
            }
        }

    @classmethod
    def union(cls, second_table_name: str, second_pipeline: list):
        # this is the SQL UNION operator
        return {
            "$unionWith": {
                "coll": second_table_name,
                "pipeline": second_pipeline
            }
        }

    @classmethod
    def sort(cls, field: str, sort_order: int) -> dict:
        # this is the SQL ORDER BY operator
        return {
            "$sort": {
                field: sort_order
            }
        }

    @classmethod
    def sort_many(cls, map_field_order: dict) -> dict:
        # this is the SQL ORDER BY operator
        return {
            "$sort": map_field_order
        }

    @classmethod
    def limit(cls, nb: int) -> dict:
        # this is the SQL LIMIT operator
        return {
            "$limit": nb
        }

    @classmethod
    def group_by(cls, group_key: dict | list | str | None, groups: list) -> dict:
        # this is the SQL GROUP BY operator
        """
        Compute group by (on one or many fields, with one or many operators)
        :param group_key: The $group stage separates documents into groups according to a "group key". The output is one document for each unique group key.
        :param groups: a list of objects for each group by to add to the query (there is only one object is one group by)
        The objects are of the form: {"name": group_by_name, "operator", the aggregation operator (min, max, avg, etc.), "field": the field on which to compute the group by
        If groups is empty, this means that we simply use the group_by operator to simulate a distinct
        :return:
        """
        query = {
            "$group": {}
        }

        query["$group"]["_id"] = group_key

        for group_by in groups:
            query["$group"][group_by["name"]] = {group_by["operator"]: group_by["field"]}
        return query

    @classmethod
    def unwind(cls, field: str) -> dict:
        return {
            "$unwind": f"${field}"
        }

    @classmethod
    def concat(cls, list_of_strings: list) -> dict:
        return {
            "$concat": list_of_strings
        }

    @classmethod
    def if_condition(cls, cond: dict, if_part: dict | str, else_part: dict | str) -> dict:
        res = {
            "$cond": {
                "if": cond,
                "then": if_part,
                "else": else_part
            }
        }
        return res

    @classmethod
    def equality(cls, field: str, value: str) -> dict:
        return {
            "$eq": [field, value]
        }

    @classmethod
    def add_fields(cls, key: str, value: str | dict) -> dict:
        return {
            "$addFields": {
                key: value
            }
        }

    @classmethod
    def set_variables(cls, variables: list) -> dict:
        return {
            "$set": {
                elem["name"]: elem["operation"] for elem in variables
            }
        }

    @classmethod
    def unset_variables(cls, variables: list) -> dict:
        return {
            "$unset": variables
        }

    @classmethod
    def from_datetime_to_isodate(cls, current_datetime: datetime) -> dict:
        return {"$date": current_datetime.strftime(THE_DATETIME_FORMAT)}

    @classmethod
    def merge(cls, table_name: str, on_attribute: str | list, when_matched: str, when_not_matched: str) -> dict:
        # append new tuples, e.g., from an aggregation pipeline, to an existing collection
        # when matched: replace|keepExisting|merge|fail|pipeline
        # when not matched: insert|discard|fail
        return {
            "$merge": {
                "into": table_name,
                "on": on_attribute,
                "whenMatched": when_matched,
                "whenNotMatched": when_not_matched
            }
        }

    @classmethod
    def write_to_table(cls, table_name: str) -> dict:
        return {
            "$out": table_name
        }

    @classmethod
    def filter_array(cls, input_array_name: str, element: str, cond: dict, limit: int) -> dict:
        return {
            "$filter": {
                "input": input_array_name,
                "as": element,
                "cond": cond,  # use element in cond to filter elements
                "limit": limit
            }
        }



from __future__ import annotations

import dataclasses
import time

from main_statistics.MainStatistics import MainStatistics
from utils.setup_logger import log


@dataclasses.dataclass()
class TimeStatistics(MainStatistics):
    stats: dict = dataclasses.field(default_factory=dict)

    def start(self, dataset: str | None, key: str):
        if dataset is None:
            dataset = "ALL"
        if dataset not in self.stats:
            self.stats[dataset] = {}
        if key not in self.stats[dataset]:
            self.stats[dataset][key] = {"start_time": 0.0, "cumulated_time": 0.0}
        self.stats[dataset][key]["start_time"] = time.time()

    def increment(self, dataset: str | None, key: str):
        if dataset is None:
            dataset = "ALL"
        if dataset in self.stats and key in self.stats[dataset]:
            self.stats[dataset][key]["cumulated_time"] += time.time() - self.stats[dataset][key]["start_time"]
        else:
            log.error(f"No existing timer for dataset {dataset} and key {key}")

    def count(self, value: int, dataset: str | None, key: str):
        if dataset is None:
            dataset = "ALL"
        if dataset not in self.stats:
            self.stats[dataset] = {}
        if key in self.stats[dataset]:
            self.stats[dataset][key]["count"] += value
        else:
            self.stats[dataset][key] = {}
            self.stats[dataset][key]["count"] = value



import unittest
from datetime import datetime

from constants.structure import TEST_DB_NAME
from database.Database import Database
from entities.Dataset import Dataset
from database.Execution import Execution
from database.Operators import Operators
from entities.Feature import Feature
from entities.Hospital import Hospital
from entities.OntologyResource import OntologyResource
from entities.Record import Record
from entities.Resource import Resource
from enums.HospitalNames import HospitalNames
from enums.Ontologies import Ontologies
from enums.ParameterKeys import ParameterKeys
from enums.Profile import Profile
from enums.TableNames import TableNames
from etl.Load import Load
from main_statistics.QualityStatistics import QualityStatistics
from utils.file_utils import write_in_file
from utils.setup_logger import log
from utils.test_utils import set_env_variables_from_dict


# personalized setup called at the beginning of each test
def my_setup(profile: str, create_indexes: bool) -> Load:
    # 1. as usual, create a Load object (to set up the current working directory)
    args = {
        ParameterKeys.DB_NAME: TEST_DB_NAME,
        ParameterKeys.DB_DROP: "True"
        # no need to set the metadata and data filepaths as we get only insert data that is written in temporary JSON files
    }
    set_env_variables_from_dict(env_vars=args)
    TestLoad.execution.internals_set_up()
    TestLoad.execution.file_set_up(setup_files=False)  # no need to set up the files, we get data and metadata as input
    database = Database(execution=TestLoad.execution)
    load = Load(database=database, execution=TestLoad.execution, create_indexes=create_indexes, dataset_id=99,
                profile=profile, quality_stats=QualityStatistics(record_stats=False))

    # 2. create few "fake" files in the current working directory in order to test insertion and index creation
    TestLoad.execution.current_dataset_gid = Dataset.compute_global_identifier()
    phen_features = [
        {
            Resource.IDENTIFIER_: 1,
            "ontology_resource": {
                OntologyResource.SYSTEM_: Ontologies.LOINC["url"],
                OntologyResource.CODE_: "123-456",
                OntologyResource.LABEL_: "age (Age in weeks)"
            },
            Resource.TIMESTAMP_: Operators.from_datetime_to_isodate(current_datetime=datetime.now()),
            Resource.ENTITY_TYPE_: f"{profile}{TableNames.FEATURE}"
        }, {
            Resource.IDENTIFIER_: 2,
            "ontology_resource": {
                OntologyResource.SYSTEM_: Ontologies.LOINC["url"],
                OntologyResource.CODE_: "123-457",
                OntologyResource.LABEL_: "twin (Whether the baby has a twin)"
            },
            Resource.TIMESTAMP_: Operators.from_datetime_to_isodate(current_datetime=datetime.now()),
            Resource.ENTITY_TYPE_: f"{profile}{TableNames.FEATURE}"
        }
    ]

    phen_records = [
        {
            Resource.IDENTIFIER_: 3,
            Record.VALUE_: 12,
            Record.SUBJECT_: 4,
            Record.REG_BY_: 1,
            Record.INSTANTIATES_: 2,
            Resource.DATASET_: f"{TestLoad.execution.current_dataset_gid}",
            Resource.TIMESTAMP_: Operators.from_datetime_to_isodate(current_datetime=datetime.now()),
            Resource.ENTITY_TYPE_: f"{profile}{TableNames.RECORD}"
        }
    ]

    patients = [
        {Resource.IDENTIFIER_: "test:1", Resource.TIMESTAMP_: Operators.from_datetime_to_isodate(current_datetime=datetime.now())},
        {Resource.IDENTIFIER_: "test:2", Resource.TIMESTAMP_: Operators.from_datetime_to_isodate(current_datetime=datetime.now())},
        {Resource.IDENTIFIER_: "test:3", Resource.TIMESTAMP_: Operators.from_datetime_to_isodate(current_datetime=datetime.now())}
    ]

    hospital = {Resource.IDENTIFIER_: f"{TableNames.HOSPITAL}:1", Hospital.NAME_: HospitalNames.TEST_H1}

    # 3. write them in temporary JSON files
    # we use 99 because we already have 1PhenotypicFeature1.json, and it would overwrite the json file
    # leading to inconsistencies and wrong inserts
    # insert the data that is inserted during the Transform step
    write_in_file(resource_list=phen_features, current_working_dir=TestLoad.execution.working_dir_current, table_name=TableNames.FEATURE, is_feature=True, dataset_id=99, to_json=False)
    load.database.insert_many_tuples(table_name=TableNames.FEATURE, tuples=phen_features)
    write_in_file(resource_list=[hospital], current_working_dir=TestLoad.execution.working_dir_current, table_name=TableNames.HOSPITAL, is_feature=False, dataset_id=99, to_json=False)
    load.database.insert_one_tuple(table_name=TableNames.HOSPITAL, one_tuple=hospital)
    # for other files, it will be inserted with the function load_remaining_data()
    write_in_file(resource_list=phen_records, current_working_dir=TestLoad.execution.working_dir_current, table_name=TableNames.RECORD, is_feature=False, dataset_id=99, to_json=False)
    write_in_file(resource_list=patients, current_working_dir=TestLoad.execution.working_dir_current, table_name=TableNames.PATIENT, is_feature=False, dataset_id=99, to_json=False)

    return load


class TestLoad(unittest.TestCase):
    execution = Execution()

    def test_run(self):
        pass

    def test_load_remaining_data(self):
        load = my_setup(profile=Profile.PHENOTYPIC, create_indexes=False)
        load.load_records()

        assert load.database.db[TableNames.FEATURE].count_documents(filter={}) == 2
        assert load.database.db[TableNames.RECORD].count_documents(filter={}) == 1
        assert load.database.db[TableNames.HOSPITAL].count_documents(filter={}) == 1

    def test_create_db_indexes(self):
        load = my_setup(profile=Profile.PHENOTYPIC, create_indexes=True)
        load.load_records()  # load also records (features, patients and hospital have been loaded in my_setup)
        load.create_db_indexes()

        # 1. for each table , we check that there are three indexes:
        #    - one on _id (mandatory, made by MongoDB)
        #    - one on identifier.value
        #    - one on timestamp
        log.info(TableNames.values(db=load.database))
        for table_name in TableNames.values(db=load.database):
            index_cursor = load.database.db[table_name].list_indexes()
            log.debug(f"table {table_name}, index_cursor: {index_cursor}")
            count_indexes = 0
            # indexes are of the form
            # SON([('v', 2), ('key', SON([('_id', 1)])), ('name', '_id_')])
            # SON([('v', 2), ('key', SON([('identifier.value', 1)])), ('name', 'identifier.value_1'), ('unique', True)])
            # SON([('v', 2), ('key', SON([('timestamp', 1)])), ('name', 'timestamp_1')])
            for index in index_cursor:
                index_keys = index["key"]
                log.info(index)
                log.info(index_keys)
                if len(index_keys) == 1 and ("_id" in index_keys or Resource.IDENTIFIER_ in index_keys or Resource.TIMESTAMP_ in index_keys or Resource.ENTITY_TYPE_ in index_keys):
                    # to check whether we have exactly the four indexes we expect
                    count_indexes += 1
                    # assert that only identifier is unique,
                    # timestamp is not (there may be several instances created at the same time)
                    # resource type is not either (we have several instances of the same type)
                    if Resource.IDENTIFIER_ in index_keys:
                        assert index["unique"] is True
                    else:
                        assert "unique" not in index
                else:
                    if table_name == TableNames.FEATURE:
                        # there is also a double index (ontology_resource.system and ontology_resource.code)
                        if len(index_keys) == 2 and f"{Feature.ONTO_}.{OntologyResource.SYSTEM_}" in index_keys and f"{Feature.ONTO_}.{OntologyResource.CODE_}" in index_keys:
                            count_indexes += 1
                            assert "unique" not in index
                        else:
                            assert False, f"{table_name} expects a compound index on two fields."
                    elif table_name == TableNames.RECORD:
                        if len(index_keys) == 5 and (Record.INSTANTIATES_ in index_keys and Record.REG_BY_ in index_keys and Record.SUBJECT_ in index_keys
                                and Resource.ENTITY_TYPE_ in index_keys and Record.BASE_ID_ in index_keys):
                            # this is the index created for upserts
                            count_indexes += 1
                            assert index["unique"] is True
                        elif len(index_keys) == 1:
                            if Record.INSTANTIATES_ in index_keys:
                                count_indexes += 1
                                assert "unique" not in index
                            elif Record.SUBJECT_ in index_keys:
                                count_indexes += 1
                                assert "unique" not in index
                            elif Resource.DATASET_ in index_keys:
                                count_indexes += 1
                                assert "unique" not in index
                            elif Record.REG_BY_ in index_keys:
                                count_indexes += 1
                                assert "unique" not in index
                            else:
                                assert False, f"{table_name} has an unknown index named {index_keys}."
                        else:
                            assert False, f"{table_name} has an unknown index named {index_keys}."
                    elif table_name == TableNames.DATASET:
                        if len(index_keys) == 1 and ("_id" in index_keys or Dataset.GID_ in index_keys):
                            count_indexes += 1
                            if Dataset.GID_ in index_keys:
                                assert index["unique"] is True
                            else:
                                assert "unique" not in index
                    else:
                        assert False, f"{table_name} should have no index."
            if table_name == TableNames.FEATURE:
                assert count_indexes == 5  # (_id, identifier, timestamp, entity_type, <onto.name, onto.code>)
            elif table_name == TableNames.RECORD:
                assert count_indexes == 8  # (_id, identifier, timestamp, entity_type, instantiates, has_subject, dataset, registered_by, <instantiates, has_subject, dataset, registered_by, entity type, base_id>)
            elif table_name == TableNames.DATASET:
                assert count_indexes == 4  # (_id, identifier, timestamp, global_identifier)
            elif table_name == TableNames.HOSPITAL:
                assert count_indexes == 4  # (_id, identifier, timestamp, entity_type)
            else:
                assert count_indexes == 4  # (_id, identifier, timestamp, entity_type)



import unittest

from entities.OntologyResource import OntologyResource
from enums.Ontologies import Ontologies


class TestOntologyResource(unittest.TestCase):

    FULL_CODE_1 = "422549004|patient-related identification code|"
    FULL_CODE_2 = "264275001 | Fluorescence polarization immunoassay technique |  :  250895007| Intensity  change |   "
    FULL_CODE_3 = "  365471004 |   finding of  details of   relatives  |    :247591002|  affected |=   (410515003|known present( qualifier value) |= 782964007|  genetic disease |)"
    FULL_CODE_4 = "GO:0000380"

    def test_compute_concat_codes(self):
        o1 = OntologyResource(system=Ontologies.SNOMEDCT, code=TestOntologyResource.FULL_CODE_1, label=None, quality_stats=None)
        o2 = OntologyResource(system=Ontologies.SNOMEDCT, code=TestOntologyResource.FULL_CODE_2, label=None, quality_stats=None)
        o3 = OntologyResource(system=Ontologies.SNOMEDCT, code=TestOntologyResource.FULL_CODE_3, label=None, quality_stats=None)
        o4 = OntologyResource(system=Ontologies.GENE_ONTOLOGY, code=TestOntologyResource.FULL_CODE_4, label=None, quality_stats=None)

        assert o1.code == "422549004"
        assert o2.code == "264275001:250895007"
        assert o3.code == "365471004:247591002=(410515003=782964007)"
        assert o4.code == "0000380"

    def test_compute_concat_names(self):
        o1 = OntologyResource(system=Ontologies.SNOMEDCT, code=TestOntologyResource.FULL_CODE_1, label=None, quality_stats=None)
        o2 = OntologyResource(system=Ontologies.SNOMEDCT, code=TestOntologyResource.FULL_CODE_2, label=None, quality_stats=None)
        o3 = OntologyResource(system=Ontologies.SNOMEDCT, code=TestOntologyResource.FULL_CODE_3, label=None, quality_stats=None)
        o4 = OntologyResource(system=Ontologies.GENE_ONTOLOGY, code=TestOntologyResource.FULL_CODE_4, label=None, quality_stats=None)

        assert o1.label == "Patient-related Identification code"
        assert o2.label == "Fluorescence polarization immunoassay technique:Intensity change"
        assert o3.label == "Details of relatives:Affecting=(Known present=Genetic disease)"
        assert o4.label == "alternative mRNA splicing via spliceosome"  # removed comma between splicing and via because this is a snomed operator



import os
import re

import pandas as pd
import jsonlines

from enums.TableNames import TableNames
from utils.setup_logger import log


def write_in_file(resource_list: list, current_working_dir: str, table_name: str, is_feature: bool, dataset_id: int, to_json: bool) -> None:
    if table_name not in [TableNames.PATIENT, TableNames.HOSPITAL, TableNames.TEST]:
        if is_feature:
            table_name = TableNames.FEATURE
        else:
            table_name = TableNames.RECORD
    filename = get_json_resource_file(current_working_dir=current_working_dir, table_name=table_name, dataset_id=dataset_id)
    if len(resource_list) > 0:
        with jsonlines.open(filename, "a") as data_file:
            try:
                # log.debug(f"Dumping {len(resource_list)} instances in {filename}")
                if to_json:
                    data_file.write_all([resource.to_json() for resource in resource_list])
                    # the_json_resources = [resource.to_json() for resource in resource_list]
                    # the_json_string = ujson.dumps(the_json_resources)
                else:
                    data_file.write_all(resource_list)
                    # the_json_resources = resource_list
                # here we write a JSONL, meaning that each record is on a line
                # there is not encompassing brackets (array), not commas between records
                # json_string = from_json_str_to_json_line(the_json_string)
                # log.info(json_string)
                # data_file.write(json_string)
            except Exception:
                raise ValueError(f"Could not dump the {len(resource_list)} JSON resources in the file located at {filename}.")
    else:
        log.info(f"No data when writing file {filename}.")


def get_json_resource_file(current_working_dir: str, dataset_id: int, table_name: str) -> str:
    return os.path.join(current_working_dir, f"{str(dataset_id)}{table_name}.jsonl")


def clear_file(current_working_dir: str, dataset_id: int, table_name: str) -> None:
    existing_file = get_json_resource_file(current_working_dir=current_working_dir,
                                           dataset_id=dataset_id,
                                           table_name=table_name)
    with open(existing_file, "w") as f:
        f.write("")


def read_tabular_file_as_string(filepath: str) -> pd.DataFrame:
    if filepath.endswith(".csv"):
        # leave empty cells as '' cells (they will be skipped during the Transform iteration on data values)
        # keep cells with explicit NaN values as they are (they will be converted into NaN during the Transform iteration on data values)
        # following issue #269
        return pd.read_csv(filepath, index_col=False, dtype=str, na_values=[], keep_default_na=False)
    elif filepath.endswith(".xls") or filepath.endswith(".xlsx"):
        # for Excel files, there may be several sheets, so we load all data in a single dataframe
        all_sheets = pd.read_excel(filepath, sheet_name=None, index_col=False, dtype=str, na_values=[], keep_default_na=False)
        all_sub_df = []
        for key, value in all_sheets.items():
            if key != "Legend":  # skip the sheet describing the columns
                all_sub_df.append(value)
        return pd.concat(all_sub_df, ignore_index=True, axis="rows")  # append lines, not vertically as new columns
    else:
        raise ValueError(f"The extension of the tabular file {filepath} is not recognised. Accepted extensions are .csv, .xls, and .xlsx.")


# transform a json-line file (one record per line, no comma, no encompassing brackets) to a valid stringified json (with brackets and commas)
def from_json_line_to_json_str(json_file) -> str:
    read_json = json_file.read()  # this is a JSON-by-line file, we need to append the encompassing brackets and a comma between each record
    read_json = "[" + read_json + "]"
    return re.sub("}\n\\{", "},{", read_json)


# transform a stringified json (e.g., outputed by json.dumps) into a str containing one record per line, no comma, and no encompassing brackets
def from_json_str_to_json_line(json_str) -> str:
    # be careful here: the easy solution is to replace },{ by }\n{
    # however, if there are nested elements, such pattern can appear in an element while not being the end of the record
    # therefore the regex needs to be tightened
    return json_str[1:-1].replace("},{", "}\n{")  # remove the brackets from the list and move each record on a line



from enums.EnumAsClass import EnumAsClass
from enums.MetadataColumns import MetadataColumns


class DiagnosisColumns(EnumAsClass):
    ID = MetadataColumns.normalize_name(column_name="Id")
    AFFECTED = MetadataColumns.normalize_name(column_name="Affected")
    DIAGNOSIS_NAME = MetadataColumns.normalize_name(column_name="DiagnosisName")
    ACRONYM = MetadataColumns.normalize_name(column_name="Acronym")
    ORPHANET_CODE = MetadataColumns.normalize_name(column_name="OrphanetCode")
    GENE_NAME = MetadataColumns.normalize_name(column_name="GeneName")
    INHERITANCE = MetadataColumns.normalize_name(column_name="VariantInheritance")
    CHR_NUMBER = MetadataColumns.normalize_name(column_name="ChromosomeNumber")
    ZIGOSITY = MetadataColumns.normalize_name(column_name="Zygosity")
    # for hospitals which have patients possibly affected by several diseases
    DISEASE_COUNTER = MetadataColumns.normalize_name(column_name="DiagnosisCounter")





from pandas import DataFrame

from database.Execution import Execution
from enums.HospitalNames import HospitalNames
from enums.Profile import Profile
from preprocessing.Preprocess import Preprocess
from preprocessing.PreprocessBuzziUC1 import PreprocessBuzziUC1
from preprocessing.PreprocessCovid import PreprocessCovid
from preprocessing.PreprocessHsjd import PreprocessHsjd
from preprocessing.PreprocessImgge import PreprocessImgge
from preprocessing.PreprocessKidneyCovid import PreprocessKidneyCovid
from preprocessing.PreprocessLafe import PreprocessLafe
from utils.setup_logger import log


class PreprocessingTask:
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str, all_columns_dataset: list):
        self.execution = execution
        self.data = data
        self.metadata = metadata
        self.profile = profile
        self.all_columns_dataset = all_columns_dataset

    def run(self):
        if self.execution.hospital_name == HospitalNames.IT_BUZZI_UC1:
            pp = PreprocessBuzziUC1(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile)
            pp.run()
        elif self.execution.hospital_name == HospitalNames.EXPES_COVID:
            pp = PreprocessCovid(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile)
            pp.run()
        elif self.execution.hospital_name == HospitalNames.EXPES_KIDNEY:
            pp = PreprocessKidneyCovid(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile)
            pp.run()
        elif self.execution.hospital_name == HospitalNames.RS_IMGGE:
            pp = PreprocessImgge(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile, all_columns_dataset=self.all_columns_dataset)
            pp.run()
        elif self.execution.hospital_name == HospitalNames.ES_HSJD:
            pp = PreprocessHsjd(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile)
            pp.run()
        elif self.execution.hospital_name == HospitalNames.ES_LAFE:
            pp = PreprocessLafe(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile, all_columns_dataset=self.all_columns_dataset)
            pp.run()
        else:
            pp = Preprocess(execution=self.execution, data=self.data, metadata=self.metadata, profile=self.profile)
            pp.run()

        # save the new data
        self.data = pp.data
        self.metadata = pp.metadata
        if hasattr(pp, "all_columns_dataset"):
            # only IMGGE pre-processing class has this attribute
            self.all_columns_dataset = pp.all_columns_dataset



from __future__ import annotations

import os
import re

from pandas import DataFrame

from constants.structure import DOCKER_FOLDER_DATA
from database.Execution import Execution
from enums.AccessTypes import AccessTypes
from enums.DiagnosisColumns import DiagnosisColumns
from enums.MetadataColumns import MetadataColumns
from enums.Profile import Profile
from preprocessing.Preprocess import Preprocess
from utils.api_utils import send_query_to_api, parse_json_response
from utils.file_utils import read_tabular_file_as_string
from utils.setup_logger import log


class PreprocessBuzziUC1(Preprocess):
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str):
        super().__init__(execution=execution, data=data, metadata=metadata, profile=profile)

        self.ids = []
        self.diagnosis_acronyms = []
        self.diagnosis_names = []
        self.affected_booleans = []
        self.orphanet = []
        self.gene_names = []
        self.inheritance = []
        self.chr_number = []
        self.zigosity = []
        self.diagnosis_counters = []
        self.mapping_diagnoses_infos = {}
        self.mapping_barcode_pid = {}

    def preprocess(self):
        log.info("pre-process BUZZI data")
        if self.profile == Profile.DIAGNOSIS:
            # 1. associate each disease to its information: gene, orphanet code, zigosity, etc
            transformation_df = read_tabular_file_as_string(os.path.join(DOCKER_FOLDER_DATA, "ds-transformation-table.xlsx"))
            transformation_df.rename(columns=lambda x: MetadataColumns.normalize_name(column_name=x), inplace=True)  # normalize column names
            transformation_df.rename(columns={"gene": DiagnosisColumns.GENE_NAME, "orpha_net": DiagnosisColumns.ORPHANET_CODE}, inplace=True)
            log.info(transformation_df)

            for row in transformation_df.itertuples(index=False):
                # acronym column
                acronym = row[transformation_df.columns.get_loc(DiagnosisColumns.ACRONYM)].lower().strip()
                if acronym not in self.mapping_diagnoses_infos:
                    self.mapping_diagnoses_infos[acronym] = {}
                # gene column
                gene_name = row[transformation_df.columns.get_loc(DiagnosisColumns.GENE_NAME)]
                if gene_name is not None and len(gene_name) > 0 and "," not in gene_name:
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.GENE_NAME] = gene_name
                else:
                    # there is no gene for that disease or this is a multigenic disease,
                    # we do not record this
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.GENE_NAME] = None
                # diagnosis name column
                diagnosis = row[transformation_df.columns.get_loc(DiagnosisColumns.DIAGNOSIS_NAME)]
                if diagnosis is not None and len(diagnosis) > 0:
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.DIAGNOSIS_NAME] = diagnosis
                else:
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.DIAGNOSIS_NAME] = None
                # orphanet code column
                orpha_code = row[transformation_df.columns.get_loc(DiagnosisColumns.ORPHANET_CODE)]
                if orpha_code != "":
                    code = orpha_code.replace("ORPHA:", "")
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.ORPHANET_CODE] = orpha_code
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.INHERITANCE] = PreprocessBuzziUC1.get_inheritance(diagnosis_code=code)
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.CHR_NUMBER] = PreprocessBuzziUC1.get_chromosome(diagnosis_code=code)
                else:
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.ORPHANET_CODE] = None
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.INHERITANCE] = None
                    self.mapping_diagnoses_infos[acronym][DiagnosisColumns.CHR_NUMBER] = None

                self.mapping_diagnoses_infos[acronym][DiagnosisColumns.ZIGOSITY] = None
            log.info(f"{len(self.mapping_diagnoses_infos)} acronyms")

            # 2. associate each sample barcode to the patient id
            prefix = Profile.get_prefix_for_path(filetype=Profile.PHENOTYPIC)
            df = read_tabular_file_as_string(filepath=f'{os.path.join(prefix, "screening.csv")}')  # cannot replace this by self.execution.current_filepath because it contains the diagnosis file data
            self.mapping_barcode_pid = {row[df.columns.get_loc("SampleBarcode")]: row[df.columns.get_loc("id")] for row in df.itertuples(index=False)}

            # 3. for each patient, collect the acronym and whether he is affected or a carrier
            count_affected = 0
            count_carrier = 0
            count_skipped = 0
            log.info(self.data.columns)
            for row in self.data.itertuples(index=False):
                pid = row[self.data.columns.get_loc("patient ID")]
                if row[self.data.columns.get_loc("affetto")] != "":
                    count_affected += 1
                    # the patient is affected by this disease, so we record this
                    self.record_diagnosis_for_patient(pid=pid, row=row, column="affetto")
                if row[self.data.columns.get_loc("carrier")] != "":
                    count_carrier += 1
                    # the patient is a carrier of the disease
                    # if we decide to record it, we record everything...
                    if self.execution.record_carrier_patients:
                        self.record_diagnosis_for_patient(pid=pid, row=row, column="carrier")
                    else:
                        # ...otherwise we do not record any information
                        count_skipped += 1

            log.info(f"count affected is {count_affected}, count carrier is {count_carrier}, count skipped is {count_skipped}, multi diagnosis is 11 = {count_affected+count_carrier+count_skipped+11}")
            log.info(f"{len(self.ids)} ids")
            log.info(f"{len(self.diagnosis_names)} diagnosis name")
            log.info(f"{len(self.diagnosis_acronyms)} acronyms")
            log.info(f"{len(self.affected_booleans)} booleans")
            log.info(f"{len(self.orphanet)} orphanet codes")
            log.info(f"{len(self.gene_names)} gene names")
            log.info(f"{len(self.inheritance)} inheritance names")
            log.info(f"{len(self.chr_number)} chr number")
            log.info(f"{len(self.zigosity)} zigosity")
            log.info(f"{len(self.diagnosis_counters)} diagnosis counters")

            self.data = DataFrame()
            self.data[DiagnosisColumns.ID] = self.ids
            self.data[DiagnosisColumns.DIAGNOSIS_NAME] = self.diagnosis_names
            self.data[DiagnosisColumns.ACRONYM] = self.diagnosis_acronyms
            self.data[DiagnosisColumns.AFFECTED] = self.affected_booleans
            self.data[DiagnosisColumns.ORPHANET_CODE] = self.orphanet
            self.data[DiagnosisColumns.GENE_NAME] = self.gene_names
            self.data[DiagnosisColumns.INHERITANCE] = self.inheritance
            self.data[DiagnosisColumns.CHR_NUMBER] = self.chr_number
            self.data[DiagnosisColumns.ZIGOSITY] = self.zigosity
            self.data[DiagnosisColumns.DISEASE_COUNTER] = self.diagnosis_counters
            log.info(self.data)
            log.info(self.data.iloc[0])

    def add_id(self, pid):
        if pid in self.mapping_barcode_pid:
            # the column is named "patient_id" in buzzi
            # but this actually contains sample bar codes
            self.ids.append(self.mapping_barcode_pid[pid])
        else:
            # we did not find the corresponding patient id in the mapping,
            # thus we simply write the sample bar code as it is
            self.ids.append(pid)

    def record_diagnosis_for_patient(self, pid, row, column: str):
        # column is "affetto" or "carrier"
        self.data.loc[:, column] = self.data[column].apply(lambda x: x.replace("/", "+") if "/" in x else x)
        for counter, disease in enumerate(row[self.data.columns.get_loc(column)].split("+")):
            self.add_id(pid=pid)
            acronym = disease.lower().strip()
            self.diagnosis_acronyms.append(acronym)
            self.diagnosis_counters.append(int(counter+1))  # +1 because enumerates starts at 0
            if column == "affetto":
                self.affected_booleans.append(True)
            else:
                self.affected_booleans.append(False)  # carrier
            if acronym in self.mapping_diagnoses_infos:
                self.diagnosis_names.append(self.mapping_diagnoses_infos[acronym][DiagnosisColumns.DIAGNOSIS_NAME])
                self.orphanet.append(self.mapping_diagnoses_infos[acronym][DiagnosisColumns.ORPHANET_CODE])
                self.gene_names.append(self.mapping_diagnoses_infos[acronym][DiagnosisColumns.GENE_NAME])
                self.inheritance.append(self.mapping_diagnoses_infos[acronym][DiagnosisColumns.INHERITANCE])
                self.chr_number.append(self.mapping_diagnoses_infos[acronym][DiagnosisColumns.CHR_NUMBER])
                self.zigosity.append(self.mapping_diagnoses_infos[acronym][DiagnosisColumns.ZIGOSITY])
            else:
                self.diagnosis_names.append(None)
                self.orphanet.append(None)
                self.gene_names.append(None)
                self.inheritance.append(None)
                self.chr_number.append(None)
                self.zigosity.append(None)

    @classmethod
    def get_inheritance(cls, diagnosis_code: str) -> str | None:
        url = f"https://api.orphadata.com/rd-natural_history/orphacodes/{diagnosis_code}"
        response = send_query_to_api(url=url, secret="nbarret", access_type=AccessTypes.API_KEY_IN_HEADER)
        data = parse_json_response(response)
        if "data" in data and "results" in data["data"] and "TypeOfInheritance" in data["data"]["results"]:
            inheritances = data["data"]["results"]["TypeOfInheritance"]
            if len(inheritances) >= 1:
                return data["data"]["results"]["TypeOfInheritance"][0]
        return None

    @classmethod
    def get_chromosome(cls, diagnosis_code: str) -> str | None:
        url = f"https://api.orphadata.com/rd-associated-genes/orphacodes/{diagnosis_code}"
        response = send_query_to_api(url=url, secret="nbarret", access_type=AccessTypes.API_KEY_IN_HEADER)
        data = parse_json_response(response)
        if "data" in data and "results" in data["data"] and "DisorderGeneAssociation" in data["data"]["results"]:
            associations = data["data"]["results"]["DisorderGeneAssociation"]
            if len(associations) >= 1:
                associations = associations[0]
                if "Gene" in associations and "Locus" in associations["Gene"]:
                    all_locus = associations["Gene"]["Locus"]
                    if len(all_locus) >= 1:
                        all_locus = all_locus[0]
                        if "GeneLocus" in all_locus:
                            full_chromosome_position = all_locus["GeneLocus"]
                            regex_elements = re.search(r"[0-9]+", full_chromosome_position)
                            return regex_elements.group()  # the chromosome_number is the first number in the string
        return None



import locale
from datetime import datetime

from dateutil.parser import parse

from enums.MetadataColumns import MetadataColumns


def cast_str_to_int(str_value: str):
    try:
        return int(str_value)
    except:
        if str_value.endswith(".0"):
            # this is a float value, that we can safely convert to int
            # this may happen when:
            # (a) Pandas converted into cell to float (because of NaN probably)
            # (b) or the data is really a float, that we can cast further
            return int(str_value.replace(".0", ""))
        else:
            # this was not an int value
            return None


def cast_str_to_float(str_value: str):
    try:
        return locale.atof(str_value)
    except:
        return None  # this was not a float value


def cast_str_to_boolean(str_value: str):
    # try to convert as boolean
    normalized_value = MetadataColumns.normalize_value(column_value=str_value)
    if normalized_value == "true" or normalized_value == "1" or normalized_value == "1.0" or normalized_value == "yes":
        return True
    elif normalized_value == "false" or normalized_value == "0" or normalized_value == "0.0" or normalized_value == "no":
        return False
    else:
        return None  # this was not a boolean value


def cast_str_to_datetime(str_value: str) -> datetime:
    try:
        datetime_value = parse(str_value)
        # %Y-%m-%d %H:%M:%S is the format used by default by parse (the output is always of this form)
        return datetime_value
    except:  # this may raise ValueError if this is not a date, or OverflowError in case of weird int (BUZZI data)
        # this was not a datetime value, and we signal it with None
        return None



from datetime import date, time, datetime

from database.Operators import Operators
from utils.setup_logger import log


def factory(data):
    # we cannot do the following because dataclasses are considering nested fields as json dictionaries (they are not Python classes anymore)
    # instead, we can check the names of the variables (not very good, but still better than nothing)
    # if not isinstance(value, (Database, Counter, QualityStatistics, TimeStatistics, DatabaseStatistics)):
    # if value is not None and key not in ["quality_stats", "time_stats", "database_stats", "counter", "database", "execution"]:
    # log.info(data)
    return {
        key: Operators.from_datetime_to_isodate(value) if isinstance(value, (datetime, date, time)) else value
        for (key, value) in data
        if value is not None and value != [] and value != {} and key not in ["quality_stats", "time_stats", "database_stats", "counter", "database", "execution", "client", "db", "dataset_key", "show_warning"]
    }
